"""

Shared fixtures for SCHISM tests.

This module provides reusable pytest fixtures for testing SCHISM functionality.
"""

import logging
import os
import shutil
import sys
import tempfile
import zipfile
from pathlib import Path

import pytest
import requests

from rompy.core.data import DataBlob, DataGrid
from rompy.core.filters import Filter
from rompy.core.source import SourceFile
from rompy.core.types import DatasetCoords
from rompy_schism.boundary_core import TidalDataset

# Import directly from the new implementation
from rompy_schism.grid import SCHISMGrid
from rompy_schism.vgrid import VGrid as SchismVGrid

DATA_REPO = "rom-py/rompy-test-data"
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
GITHUB_API_RELEASES = f"https://api.github.com/repos/{DATA_REPO}/releases/latest"

# Add the tests directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

# Import test utilities for logging
from test_utils.logging import configure_test_logging

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session", autouse=True)
def docker_available() -> bool:
    try:
        result = subprocess.run(
            [
                "docker",
                "info",
            ],
            capture_output=True,
            text=True,
            timeout=10,
        )
        return result.returncode == 0
    except Exception:
        return False


# Helper to detect CI environment
def is_ci_environment():
    """Detect if running in a CI environment."""
    ci_vars = [
        "CI",
        "GITHUB_ACTIONS",
        "GITLAB_CI",
        "TRAVIS",
        "CIRCLECI",
        "APPVEYOR",
        "BUILDKITE",
        "TEAMCITY_VERSION",
    ]
    return any(
        os.environ.get(var, "").lower() in ("1", "true", "yes", "on") for var in ci_vars
    )


@pytest.fixture(scope="session")
def should_skip_docker_builds():
    """Fixture to determine if Docker builds should be skipped."""
    if os.environ.get("SKIP_DOCKER_BUILDS", "").lower() in ("1", "true", "yes", "on"):
        return True
    if os.environ.get("ROMPY_SKIP_DOCKER_BUILDS", "").lower() in (
        "1",
        "true",
        "yes",
        "on",
    ):
        return True
    if os.environ.get("ROMPY_ENABLE_DOCKER_IN_CI", "").lower() not in (
        "1",
        "true",
        "yes",
        "on",
    ):
        return is_ci_environment()
    return False


def pytest_addoption(parser):
    """Add command-line options for pytest."""
    parser.addoption(
        "--run-slow",
        action="store_true",
        default=False,
        help="Run slow tests",
    )
    parser.addoption(
        "--rompy-log-level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level for ROMPY tests",
    )


def pytest_configure(config):
    """Configure pytest with plugins and settings, and ensure test data is present."""
    import logging

    # Get log level from command line or use default
    log_level_str = config.getoption("--rompy-log-level")
    log_level = getattr(logging, log_level_str)

    # Configure logging for tests
    configure_test_logging(level=log_level_str)


def download_and_extract_data():
    print("Downloading test data from rompy-test-data repo...")
    # Get latest release info
    resp = requests.get(GITHUB_API_RELEASES)
    resp.raise_for_status()
    release = resp.json()
    # Find the zipball URL
    zip_url = release["zipball_url"]
    # Download the zip
    with tempfile.TemporaryDirectory() as tmpdir:
        zip_path = os.path.join(tmpdir, "data.zip")
        with requests.get(zip_url, stream=True) as r:
            r.raise_for_status()
            with open(zip_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        # Extract only the data/ directory
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            # Find the top-level directory in the zip
            top_level = zip_ref.namelist()[0].split("/")[0]
            # Extract data/ to DATA_DIR
            for member in zip_ref.namelist():
                if member.startswith(f"{top_level}/data/"):
                    zip_ref.extract(member, tmpdir)
            src_data_dir = os.path.join(tmpdir, top_level, "data")
            if os.path.exists(DATA_DIR):
                shutil.rmtree(DATA_DIR)
            shutil.copytree(src_data_dir, DATA_DIR)
    print("Test data downloaded and extracted.")


# Only download if data dir is missing or empty
if not os.path.exists(DATA_DIR) or not os.listdir(DATA_DIR):
    try:
        download_and_extract_data()
    except Exception as e:
        print(f"Failed to download test data: {e}", file=sys.stderr)
        sys.exit(1)
    # --- END: Automatic test data download ---


@pytest.fixture(scope="session", autouse=True)
def setup_logging():
    """Set up logging for all tests.

    This fixture runs once per test session and ensures that logging is properly
    configured for all tests.
    """
    # Import here to avoid circular imports
    from test_utils.logging import configure_test_logging

    # Configure logging with default settings
    configure_test_logging()

    # Return a function to reconfigure logging if needed
    return configure_test_logging


@pytest.fixture
def test_files_dir():
    """Return path to test files directory (old structure)."""
    return Path(__file__).parent / "data" / "schism"


@pytest.fixture
def hycom_path(test_files_dir):
    """Get the path to HYCOM data."""
    hycomdata = test_files_dir / "hycom.nc"
    if not hycomdata.exists():
        from tests.utils import download_hycom

        logging.info("Hycom test data not found, downloading...")
        logging.info("This may take a while...only has to be done once.")
        download_hycom(dest=test_files_dir, hgrid=test_files_dir / "hgrid.gr3")

    return str(hycomdata)


@pytest.fixture
def hgrid_path(test_files_dir):
    """Return path to a test hgrid.gr3 file."""
    # Look for hgrid.gr3 files in the test data directory
    potential_files = list(test_files_dir.glob("**/hgrid.gr3"))
    if potential_files:
        return potential_files[0]

    # Fall back to the hgrid_20kmto60km file if no hgrid.gr3 found
    fallback = Path(__file__).parent / "hgrid_20kmto60km_rompyschism_testing.gr3"
    if fallback.exists():
        return fallback

    pytest.skip("No suitable hgrid file found for testing")
    return None


@pytest.fixture
def grid2d(test_files_dir):
    """Return a 2D SCHISM grid for testing."""
    # Create a grid with DataBlob for hgrid
    grid = SCHISMGrid(
        hgrid=DataBlob(source=test_files_dir / "hgrid.gr3"),
        drag=1.0,
    )
    return grid


@pytest.fixture
def grid3d(test_files_dir):
    """Return a 3D SCHISM grid with vgrid for testing."""
    vgrid_path = test_files_dir / "vgrid.in"
    if vgrid_path.exists():
        vgrid = DataBlob(source=vgrid_path)
    else:
        # Create a basic vertical grid with default values
        vgrid = SchismVGrid(
            ivcor=2, nvrt=20, h_s=20.0, theta_b=0.5, theta_f=5.0  # LSC2
        )

    # Create the grid with both hgrid and vgrid
    grid = SCHISMGrid(
        hgrid=DataBlob(source=test_files_dir / "hgrid.gr3"),
        vgrid=vgrid,
        drag=1.0,
    )

    return grid


@pytest.fixture
def grid_atmos_source(test_files_dir):
    """Create a source for atmospheric data."""
    return DataGrid(
        source=SourceFile(uri=str(test_files_dir / "air_1.nc")),
        coords=DatasetCoords(t="time", x="lon", y="lat"),
        variables=[
            "air_pressure",
            "air_temperature",
            "specific_humidity",
            "eastward_wind",
            "northward_wind",
        ],
        buffer=0.1,
        filter=Filter(),
        crop_data=True,
    )


@pytest.fixture
def hycom_bnd_elev(test_files_dir, hycom_path):
    """Create a 2D hydrodynamic boundary source."""
    return DataGrid(
        source=SourceFile(uri=hycom_path),
        coords=DatasetCoords(t="time", x="lon", y="lat"),
        variables=["surf_el"],
        buffer=0.1,
        filter=Filter(),
        crop_data=True,
    )


@pytest.fixture
def hycom_bnd_vel(test_files_dir, hycom_path):
    """Create a 2D hydrodynamic boundary source."""
    return DataGrid(
        source=SourceFile(uri=hycom_path),
        coords=DatasetCoords(t="time", x="lon", y="lat"),
        variables=["water_u", "water_v"],
        buffer=0.1,
        filter=Filter(),
        crop_data=True,
    )


@pytest.fixture
def hycom_bnd_temp_3d(test_files_dir, hycom_path):
    """Create a 3D temperature boundary source."""
    return DataGrid(
        source=SourceFile(uri=hycom_path),
        coords=DatasetCoords(t="time", x="lon", y="lat", z="depth"),
        variables=["water_temp"],
        buffer=0.1,
        filter=Filter(),
        crop_data=True,
    )


@pytest.fixture
def hycom_bnd2d(test_files_dir, hycom_path):
    """Create a 3D temperature boundary source."""
    return DataGrid(
        source=SourceFile(uri=hycom_path),
        coords=DatasetCoords(t="time", x="lon", y="lat", z="depth"),
    )


@pytest.fixture
def tidal_data_files(test_files_dir):
    """Return paths to tidal elevation and velocity files for testing."""
    tidal_database = test_files_dir / "tides"
    if not (tidal_database / "oceanum-atlas" / "grid_tpxo9_atlas_30_v2.nc").exists():
        if (tidal_database / "oceanum-atlas.tar.gz").exists():
            import tarfile

            logger.info(f"Unpacking {tidal_database / 'oceanum-atlas.tar.gz'}")
            with tarfile.open(tidal_database / "oceanum-atlas.tar.gz") as tar:
                tar.extractall(path=tidal_database)
    return tidal_database


@pytest.fixture
def tidal_dataset(tidal_data_files):
    """Return a tidal dataset instance for testing."""

    return TidalDataset(
        tidal_database=tidal_data_files,
        constituents=["M2", "S2"],
        tidal_model="OCEANUM-atlas",
    )


@pytest.fixture
def mock_tidal_data():
    """Create mock tidal data for testing."""
    import numpy as np

    # Mock data for testing - enough for any boundary size
    # For elevation: [amplitude, phase]
    # For velocity: [u_amplitude, u_phase, v_amplitude, v_phase]
    def mock_data(self, lons, lats, tnames, data_type="h"):
        num_nodes = len(lons)
        num_constituents = len(tnames)
        if data_type == "h":  # Elevation
            arr = np.zeros((num_nodes, num_constituents, 2))
            arr[..., 0] = 0.5  # amplitude
            arr[..., 1] = 45.0  # phase
            return arr
        elif data_type == "uv":  # Velocity
            arr = np.zeros((num_nodes, num_constituents, 4))
            arr[..., 0] = 0.1  # u_amp
            arr[..., 1] = 30.0  # u_phase
            arr[..., 2] = 0.2  # v_amp
            arr[..., 3] = 60.0  # v_phase
            return arr
        else:
            raise ValueError(f"Unknown data type: {data_type}")

    return mock_data



"""Tests for the atmospheric forcing (sflux) plotting methods in SCHISMDataSflux."""

from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr

pytest.importorskip("rompy_schism")


# Define the location of test files
HERE = Path(__file__).parent


@pytest.fixture
def test_air_dataset():
    """Create a sample air dataset for testing sflux air plotting."""
    # Create a sample grid and time dimension
    lon = np.linspace(-75, -70, 20)
    lat = np.linspace(35, 40, 15)
    times = pd.date_range(start=datetime.now(), periods=5, freq="1D").to_pydatetime()

    # Create 2D meshgrid for spatial data
    lon_grid, lat_grid = np.meshgrid(lon, lat)

    # Create some sample data
    num_times = len(times)
    num_lats = len(lat)
    num_lons = len(lon)

    # Air pressure at mean sea level
    prmsl = np.zeros((num_times, num_lats, num_lons))
    # Wind components
    uwind = np.zeros((num_times, num_lats, num_lons))
    vwind = np.zeros((num_times, num_lats, num_lons))
    # Air temperature
    stmp = np.zeros((num_times, num_lats, num_lons))
    # Specific humidity
    spfh = np.zeros((num_times, num_lats, num_lons))

    # Fill with meaningful data
    for t in range(num_times):
        # Center of low pressure system moving across the domain
        center_lon = -75 + t * 1.0
        center_lat = 35 + t * 0.8

        for i in range(num_lats):
            for j in range(num_lons):
                # Distance from pressure center
                dist = np.sqrt((lon[j] - center_lon) ** 2 + (lat[i] - center_lat) ** 2)

                # Pressure field (low in center, higher outward)
                prmsl[t, i, j] = 101300 - 2000 * np.exp(-0.5 * dist**2)

                # Wind field (circular around low pressure)
                dx = lon[j] - center_lon
                dy = lat[i] - center_lat
                wind_speed = 10 * (1 - np.exp(-0.3 * dist))
                angle = np.arctan2(dy, dx) + np.pi / 2  # Counter-clockwise around low
                uwind[t, i, j] = wind_speed * np.cos(angle)
                vwind[t, i, j] = wind_speed * np.sin(angle)

                # Temperature decreases with latitude and time
                stmp[t, i, j] = 25 - 0.2 * (lat[i] - 35) - 0.5 * t

                # Humidity has a similar but weaker pattern to pressure
                spfh[t, i, j] = 0.015 - 0.005 * np.exp(-0.5 * dist**2)

    # Create xarray dataset
    ds = xr.Dataset(
        {
            "prmsl": (["time", "lat", "lon"], prmsl),
            "uwind": (["time", "lat", "lon"], uwind),
            "vwind": (["time", "lat", "lon"], vwind),
            "stmp": (["time", "lat", "lon"], stmp),
            "spfh": (["time", "lat", "lon"], spfh),
        },
        coords={
            "time": times,
            "lat": lat,
            "lon": lon,
        },
    )

    # Add unit attributes
    ds["prmsl"].attrs["units"] = "Pa"
    ds["uwind"].attrs["units"] = "m/s"
    ds["vwind"].attrs["units"] = "m/s"
    ds["stmp"].attrs["units"] = "degC"
    ds["spfh"].attrs["units"] = "kg/kg"

    return ds


@pytest.fixture
def test_sflux_data(test_air_dataset):
    """Create sample sflux data for testing without using SfluxAir or SCHISMDataSflux."""

    # Create a simple class that just returns the dataset
    class SimpleSfluxDataSource:
        def __init__(self, dataset):
            self.air_1 = self.AirContainer(dataset)

        class AirContainer:
            def __init__(self, dataset):
                self.dataset = dataset

    # Return a simple container with the air dataset
    return SimpleSfluxDataSource(test_air_dataset)


def test_plot_sflux_spatial(grid2d, test_sflux_data):
    """Test plotting of sflux spatial fields."""
    import matplotlib.pyplot as plt
    import numpy as np

    # Get the raw dataset to avoid validation issues
    air_ds = test_sflux_data.air_1.dataset

    # Test plotting air pressure field
    fig, ax = plt.subplots(figsize=(10, 8))

    # Get grid for contour plot
    lons = air_ds.lon.values
    lats = air_ds.lat.values
    lon_grid, lat_grid = np.meshgrid(lons, lats)

    # Plot contour of pressure at first time step
    cs = ax.contourf(lon_grid, lat_grid, air_ds.prmsl[0, :, :] / 100, cmap="viridis")
    plt.colorbar(cs, ax=ax, label="Pressure (hPa)")

    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.set_title("Air Pressure at Sea Level")

    assert fig is not None
    assert ax is not None

    # Test with wind vectors
    fig, ax = plt.subplots(figsize=(10, 8))

    # Plot wind vectors on a subset of the grid for clarity
    skip = 2  # Skip every other grid point
    ax.quiver(
        lon_grid[::skip, ::skip],
        lat_grid[::skip, ::skip],
        air_ds.uwind[0, ::skip, ::skip],
        air_ds.vwind[0, ::skip, ::skip],
        scale=100,
    )

    # Add colorbar as background showing wind speed
    wind_speed = np.sqrt(air_ds.uwind[0, :, :] ** 2 + air_ds.vwind[0, :, :] ** 2)
    cs = ax.contourf(lon_grid, lat_grid, wind_speed, cmap="RdBu_r", alpha=0.5)
    plt.colorbar(cs, ax=ax, label="Wind Speed (m/s)")

    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.set_title("Wind Field")

    assert fig is not None
    assert ax is not None


def test_plot_sflux_timeseries(test_sflux_data):
    """Test plotting of sflux time series."""
    import matplotlib.pyplot as plt

    # Get the raw dataset to avoid validation issues
    air_ds = test_sflux_data.air_1.dataset

    # Test with temperature time series at a specific point
    fig, ax = plt.subplots(figsize=(10, 6))

    # Choose a grid point for the time series
    lat_idx, lon_idx = 5, 10

    # Plot temperature time series
    ax.plot(air_ds.time, air_ds.stmp[:, lat_idx, lon_idx], "r-", marker="o")
    ax.set_xlabel("Time")
    ax.set_ylabel("Temperature (°C)")
    ax.set_title(
        f"Air Temperature at Point ({air_ds.lon[lon_idx]:.2f}°E, {air_ds.lat[lat_idx]:.2f}°N)"
    )
    ax.grid(True)

    assert fig is not None

    # Test with wind speed time series
    fig, ax = plt.subplots(figsize=(10, 6))

    # Calculate wind speed
    wind_speed = np.sqrt(
        air_ds.uwind[:, lat_idx, lon_idx] ** 2 + air_ds.vwind[:, lat_idx, lon_idx] ** 2
    )

    # Plot wind speed time series
    ax.plot(air_ds.time, wind_speed, "b-", marker="s")
    ax.set_xlabel("Time")
    ax.set_ylabel("Wind Speed (m/s)")
    ax.set_title(
        f"Wind Speed at Point ({air_ds.lon[lon_idx]:.2f}°E, {air_ds.lat[lat_idx]:.2f}°N)"
    )
    ax.grid(True)

    assert fig is not None


def test_sflux_plotting_workflow(grid2d, test_sflux_data):
    """Test a complete workflow with multiple sflux plots."""
    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib.gridspec import GridSpec

    # Get the raw dataset
    air_ds = test_sflux_data.air_1.dataset

    # Create a figure with multiple subplots for a complete workflow
    fig = plt.figure(figsize=(15, 12))
    gs = GridSpec(2, 2, figure=fig)

    # Get mesh grid for spatial plots
    lons = air_ds.lon.values
    lats = air_ds.lat.values
    lon_grid, lat_grid = np.meshgrid(lons, lats)

    # 1. Air pressure plot
    ax1 = fig.add_subplot(gs[0, 0])
    cs1 = ax1.contourf(lon_grid, lat_grid, air_ds.prmsl[0, :, :] / 100, cmap="viridis")
    plt.colorbar(cs1, ax=ax1, label="Pressure (hPa)")
    ax1.set_xlabel("Longitude")
    ax1.set_ylabel("Latitude")
    ax1.set_title("Air Pressure at Sea Level")

    # 2. Wind vectors plot
    ax2 = fig.add_subplot(gs[0, 1])
    skip = 2  # Skip grid points for clarity
    ax2.quiver(
        lon_grid[::skip, ::skip],
        lat_grid[::skip, ::skip],
        air_ds.uwind[0, ::skip, ::skip],
        air_ds.vwind[0, ::skip, ::skip],
        scale=100,
    )

    # Background showing wind speed
    wind_speed = np.sqrt(air_ds.uwind[0, :, :] ** 2 + air_ds.vwind[0, :, :] ** 2)
    cs2 = ax2.contourf(lon_grid, lat_grid, wind_speed, cmap="RdBu_r", alpha=0.5)
    plt.colorbar(cs2, ax=ax2, label="Wind Speed (m/s)")

    ax2.set_xlabel("Longitude")
    ax2.set_ylabel("Latitude")
    ax2.set_title("Wind Field")

    # 3. Temperature time series
    ax3 = fig.add_subplot(gs[1, 0])
    lat_idx, lon_idx = 7, 10  # Choose a point
    ax3.plot(air_ds.time, air_ds.stmp[:, lat_idx, lon_idx], "r-", marker="o")
    ax3.set_xlabel("Time")
    ax3.set_ylabel("Temperature (°C)")
    ax3.set_title("Air Temperature Time Series")
    ax3.grid(True)

    # 4. Pressure field at a different time
    ax4 = fig.add_subplot(gs[1, 1])
    time_idx = 2  # Different time step
    cs4 = ax4.contourf(
        lon_grid, lat_grid, air_ds.prmsl[time_idx, :, :] / 100, cmap="viridis"
    )
    plt.colorbar(cs4, ax=ax4, label="Pressure (hPa)")
    ax4.set_xlabel("Longitude")
    ax4.set_ylabel("Latitude")
    ax4.set_title(f"Air Pressure at Time {time_idx + 1}")

    plt.tight_layout()

    assert fig is not None



# This file was auto generated from a SCHISM namelist file on 2025-01-24.

from typing import Optional

from pydantic import Field, field_validator

from rompy_schism.namelists.basemodel import NamelistBaseModel


class Mice_in(NamelistBaseModel):
    ice_tests: Optional[int] = Field(
        0, description="Flag for box test. 0 indicates no box test."
    )
    ihot_mice: Optional[int] = Field(
        1,
        description="Start mode for ice model. 0: cold start, 1: restart, 2: hotstart from HYCOM.",
    )
    ice_advection: Optional[int] = Field(
        6,
        description="Ice advection scheme. 3: upwind, 4: center-difference, 5: tvd, 6: tvd-up, 7: TVD_Casulli.",
    )
    ice_therm_on: Optional[int] = Field(
        1, description="Flag for ice thermodynamics. 1: on, 0: off."
    )
    ievp: Optional[int] = Field(
        2,
        description="Elastic-Viscous-Plastic (EVP) model selection. 1: EVP, 2: modified EVP (mEVP).",
    )
    ice_cutoff: Optional[float] = Field(
        0.001,
        description="Cut-off thickness [m] or fraction for ice. No ice velocity if ice thickness <= ice_cutoff.",
    )
    evp_rheol_steps: Optional[int] = Field(
        500, description="Number of subcycling steps in EVP."
    )
    mevp_rheol_steps: Optional[int] = Field(
        500, description="Number of iterations in modified EVP (mEVP)."
    )
    delta_min: Optional[float] = Field(
        1e-11,
        description="Limit for minimum divergence (1/s). Used in both VP and EVP.",
    )
    theta_io: Optional[float] = Field(
        0.0, description="Ice/ocean rotation angle in degrees."
    )
    mevp_alpha1: Optional[float] = Field(
        200.0, description="Constant used in mEVP for constitutive equation."
    )
    mevp_alpha2: Optional[float] = Field(
        200.0, description="Constant used in mEVP for momentum equation."
    )
    pstar: Optional[float] = Field(
        27500.0, description="Ice strength parameter [N/m^2]."
    )
    ellipse: Optional[float] = Field(
        2.0, description="Ellipticity parameter for ice rheology."
    )
    c_pressure: Optional[float] = Field(
        20.0, description="Ice pressure coefficient [-]."
    )
    niter_fct: Optional[int] = Field(
        3,
        description="Number of iterations in higher-order solve for FCT (Flux-Corrected Transport).",
    )
    ice_gamma_fct: Optional[float] = Field(
        0.25, description="Smoothing parameter for FCT."
    )
    h_ml0: Optional[float] = Field(
        0.1, description="Initial ocean mixed layer depth [m]."
    )
    salt_ice: Optional[float] = Field(5.0, description="Salinity for ice [PSU].")
    salt_water: Optional[float] = Field(34.0, description="Salinity for water [PSU].")

    @field_validator("ice_tests")
    @classmethod
    def validate_ice_tests(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("ice_tests must be 0 or 1")
        return v

    @field_validator("ihot_mice")
    @classmethod
    def validate_ihot_mice(cls, v):
        if not isinstance(v, int) or v not in [0, 1, 2]:
            raise ValueError("ihot_mice must be 0, 1, or 2")
        return v

    @field_validator("ice_advection")
    @classmethod
    def validate_ice_advection(cls, v):
        if not isinstance(v, int) or v not in range(3, 8):
            raise ValueError("ice_advection must be an integer between 3 and 7")
        return v

    @field_validator("ice_therm_on")
    @classmethod
    def validate_ice_therm_on(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("ice_therm_on must be 0 or 1")
        return v

    @field_validator("ievp")
    @classmethod
    def validate_ievp(cls, v):
        if not isinstance(v, int) or v not in [1, 2]:
            raise ValueError("ievp must be 1 or 2")
        return v

    @field_validator("ice_cutoff")
    @classmethod
    def validate_ice_cutoff(cls, v):
        if not isinstance(v, float) or v < 0:
            raise ValueError("ice_cutoff must be a non-negative float")
        return v

    @field_validator("evp_rheol_steps")
    @classmethod
    def validate_evp_rheol_steps(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("evp_rheol_steps must be a positive integer")
        return v

    @field_validator("mevp_rheol_steps")
    @classmethod
    def validate_mevp_rheol_steps(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("mevp_rheol_steps must be a positive integer")
        return v

    @field_validator("delta_min")
    @classmethod
    def validate_delta_min(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("delta_min must be a positive float")
        return v

    @field_validator("theta_io")
    @classmethod
    def validate_theta_io(cls, v):
        if not isinstance(v, float) or v < 0 or v >= 360:
            raise ValueError("theta_io must be a float between 0 and 360")
        return v

    @field_validator("mevp_alpha1")
    @classmethod
    def validate_mevp_alpha1(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("mevp_alpha1 must be a positive float")
        return v

    @field_validator("mevp_alpha2")
    @classmethod
    def validate_mevp_alpha2(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("mevp_alpha2 must be a positive float")
        return v

    @field_validator("pstar")
    @classmethod
    def validate_pstar(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("pstar must be a positive float")
        return v

    @field_validator("ellipse")
    @classmethod
    def validate_ellipse(cls, v):
        if not isinstance(v, float) or v <= 1:
            raise ValueError("ellipse must be a float greater than 1")
        return v

    @field_validator("c_pressure")
    @classmethod
    def validate_c_pressure(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("c_pressure must be a positive float")
        return v

    @field_validator("niter_fct")
    @classmethod
    def validate_niter_fct(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("niter_fct must be a positive integer")
        return v

    @field_validator("ice_gamma_fct")
    @classmethod
    def validate_ice_gamma_fct(cls, v):
        if not isinstance(v, float) or v < 0 or v > 1:
            raise ValueError("ice_gamma_fct must be a float between 0 and 1")
        return v

    @field_validator("h_ml0")
    @classmethod
    def validate_h_ml0(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("h_ml0 must be a positive float")
        return v

    @field_validator("salt_ice")
    @classmethod
    def validate_salt_ice(cls, v):
        if not isinstance(v, float) or v < 0:
            raise ValueError("salt_ice must be a non-negative float")
        return v

    @field_validator("salt_water")
    @classmethod
    def validate_salt_water(cls, v):
        if not isinstance(v, float) or v < 0:
            raise ValueError("salt_water must be a non-negative float")
        return v


class Mice(NamelistBaseModel):
    mice_in: Optional[Mice_in] = Field(default_factory=Mice_in)



"""
Test for GR3 file generation using PyLibs implementation.
"""

import tempfile
from pathlib import Path

import pytest

from rompy_schism.grid import GR3Generator


# Skip this test if required files don't exist
def get_test_hgrid():
    """Find a suitable hgrid.gr3 file for testing"""
    # Try to find a test grid in the test data directory
    test_data_dir = Path(__file__).parent / "data"

    # Look for hgrid.gr3 files in the test data directory
    potential_files = list(test_data_dir.glob("**/hgrid.gr3"))

    if potential_files:
        return potential_files[0]

    # If no file found, skip the test
    pytest.skip("No hgrid.gr3 file found for testing")
    return None


@pytest.mark.parametrize(
    "gr3_type,value",
    [
        ("drag", 0.0025),
        ("diffmin", 1.0),
        ("diffmax", 1.0),
        ("watertype", 1),
        ("albedo", 0.1),
        ("windrot_geo2proj", 0.0),
    ],
)
def test_gr3_generation(gr3_type, value):
    """Test that GR3 file can be generated properly with PyLibs implementation"""
    hgrid_path = get_test_hgrid()

    if hgrid_path is None:
        pytest.skip("No hgrid.gr3 file found for testing")

    # Create temporary directory for output
    with tempfile.TemporaryDirectory() as tmpdir:
        # Create generator
        generator = GR3Generator(hgrid=hgrid_path, gr3_type=gr3_type, value=value)

        # Generate GR3 file
        output_file = generator.generate(tmpdir)

        # Check that file was created
        assert output_file.exists(), f"Failed to generate {gr3_type}.gr3 file"


def test_generate_all_gr3_files():
    """
    Generate all GR3 files at once to verify the implementation works correctly.
    """
    hgrid_path = get_test_hgrid()
    if hgrid_path is None:
        pytest.skip("No hgrid.gr3 file found for testing")

    # Output directory
    with tempfile.TemporaryDirectory() as tmpdir:
        output_dir = Path(tmpdir)

        # Generate all GR3 file types
        gr3_types = {
            "drag": 0.0025,
            "diffmin": 1.0e-6,
            "diffmax": 1.0,
            "watertype": 1,
            "albedo": 0.15,
            "windrot_geo2proj": 0.0,
        }

        for gr3_type, value in gr3_types.items():
            generator = GR3Generator(hgrid=hgrid_path, gr3_type=gr3_type, value=value)
            output_file = generator.generate(output_dir)
            assert output_file.exists(), f"Failed to generate {gr3_type}.gr3 file"


if __name__ == "__main__":
    # When run directly, generate test GR3 files
    hgrid_path = get_test_hgrid()
    if hgrid_path:
        output_dir = Path(__file__).parent / "gr3_files"
        output_dir.mkdir(exist_ok=True, parents=True)

        # Generate all GR3 file types
        gr3_types = {
            "drag": 0.0025,
            "diffmin": 1.0e-6,
            "diffmax": 1.0,
            "watertype": 1,
            "albedo": 0.15,
            "windrot_geo2proj": 0.0,
        }

        for gr3_type, value in gr3_types.items():
            print(f"Generating {gr3_type}.gr3...")
            generator = GR3Generator(hgrid=hgrid_path, gr3_type=gr3_type, value=value)
            output_file = generator.generate(output_dir)
            print(f"Generated {output_file}")



from datetime import datetime
from pathlib import Path

import numpy as np
import pytest

from rompy_schism.bctides import Bctides
from rompy_schism.boundary_core import TidalBoundary  # Backward compatibility alias
from rompy_schism.boundary_core import (
    BoundaryConfig,
    ElevationType,
    TracerType,
    VelocityType,
    create_hybrid_boundary,
    create_nested_boundary,
    create_river_boundary,
    create_tidal_boundary,
)


class TestBoundaryConfig:
    """Tests for the BoundaryConfig class."""

    def test_init_default(self):
        """Test initialization with default values."""
        config = BoundaryConfig()

        assert config.elev_type == ElevationType.NONE
        assert config.vel_type == VelocityType.NONE
        assert config.temp_type == TracerType.NONE
        assert config.salt_type == TracerType.NONE

        assert config.ethconst is None
        assert config.vthconst is None
        assert config.tthconst is None
        assert config.sthconst is None

        assert config.inflow_relax == 0.5
        assert config.outflow_relax == 0.1

        assert config.tobc == 1.0
        assert config.sobc == 1.0

    def test_init_custom(self):
        """Test initialization with custom values."""
        config = BoundaryConfig(
            elev_type=ElevationType.HARMONIC,
            vel_type=VelocityType.HARMONIC,
            temp_type=TracerType.CONSTANT,
            salt_type=TracerType.CONSTANT,
            ethconst=1.0,
            vthconst=-100.0,
            tthconst=15.0,
            sthconst=35.0,
            inflow_relax=0.8,
            outflow_relax=0.2,
        )

        assert config.elev_type == ElevationType.HARMONIC
        assert config.vel_type == VelocityType.HARMONIC
        assert config.temp_type == TracerType.CONSTANT
        assert config.salt_type == TracerType.CONSTANT

        assert config.ethconst == 1.0
        assert config.vthconst == -100.0
        assert config.tthconst == 15.0
        assert config.sthconst == 35.0

        assert config.inflow_relax == 0.8
        assert config.outflow_relax == 0.2

        assert config.tobc == 1.0
        assert config.sobc == 1.0

    def test_flather_validation(self):
        """Test validation for Flather boundary conditions."""
        config = BoundaryConfig(
            elev_type=ElevationType.NONE, vel_type=VelocityType.FLATHER
        )

        # Set eta_mean and vn_mean for Flather boundary
        num_nodes = 5  # Just for testing
        config.eta_mean = np.ones(num_nodes) * 0.1
        config.vn_mean = np.ones(num_nodes) * 0.05

        # Validate that eta_mean and vn_mean are properly set
        assert isinstance(config.eta_mean, np.ndarray)
        assert isinstance(config.vn_mean, np.ndarray)
        assert len(config.eta_mean) == num_nodes
        assert len(config.vn_mean) == num_nodes


@pytest.fixture
def sample_grid_path():
    """Return path to a test grid file."""
    grid_path = Path(__file__).parent / "hgrid_20kmto60km_schism_testing.gr3"
    if not grid_path.exists():
        pytest.skip("Test grid file not found")
    return str(grid_path)


@pytest.fixture
def sample_tidal_files():
    """Return paths to tidal data files."""
    test_data_dir = Path(__file__).parent / "data" / "schism" / "tpxo9-neaus"
    elev_file = test_data_dir / "h_m2s2n2.nc"
    vel_file = test_data_dir / "u_m2s2n2.nc"

    if not elev_file.exists() or not vel_file.exists():
        pytest.skip("Tidal data files not found")

    return {"elevations": str(elev_file), "velocities": str(vel_file)}


class TestTidalBoundary:
    """Tests for the TidalBoundary class."""

    def test_init(self, sample_grid_path):
        """Test initialization with a grid file."""
        boundary = TidalBoundary(
            grid_path=sample_grid_path, constituents=["M2", "S2", "N2"]
        )

        assert boundary.constituents == ["M2", "S2", "N2"]
        assert boundary.tidal_database == "tpxo"
        assert boundary.grid is not None

    def test_set_boundary_config(self, sample_grid_path):
        """Test setting boundary configuration."""
        boundary = TidalBoundary(grid_path=sample_grid_path)

        config = BoundaryConfig(
            elev_type=ElevationType.HARMONIC, vel_type=VelocityType.HARMONIC
        )

        boundary.set_boundary_config(0, config)
        assert 0 in boundary.boundary_configs
        assert boundary.boundary_configs[0].elev_type == ElevationType.HARMONIC
        assert boundary.boundary_configs[0].vel_type == VelocityType.HARMONIC

    def test_set_boundary_type(self, sample_grid_path):
        """Test setting boundary type."""
        boundary = TidalBoundary(grid_path=sample_grid_path)

        boundary.set_boundary_type(
            0,
            elev_type=ElevationType.HARMONIC,
            vel_type=VelocityType.HARMONIC,
            temp_type=TracerType.NONE,
            salt_type=TracerType.NONE,
        )

        assert 0 in boundary.boundary_configs
        assert boundary.boundary_configs[0].elev_type == ElevationType.HARMONIC
        assert boundary.boundary_configs[0].vel_type == VelocityType.HARMONIC
        assert boundary.boundary_configs[0].temp_type == TracerType.NONE
        assert boundary.boundary_configs[0].salt_type == TracerType.NONE

    def test_get_flags_list(self, sample_grid_path):
        """Test getting flags list from configurations."""
        boundary = TidalBoundary(grid_path=sample_grid_path)

        # Add configurations for multiple boundaries
        boundary.set_boundary_type(0, ElevationType.HARMONIC, VelocityType.HARMONIC)
        boundary.set_boundary_type(
            1,
            ElevationType.CONSTANT,
            VelocityType.CONSTANT,
            vthconst=-100.0,
            ethconst=1.0,
        )
        boundary.set_boundary_type(3, ElevationType.NONE, VelocityType.FLATHER)

        flags = boundary.get_flags_list()

        assert len(flags) == 4  # Should include boundary indices 0, 1, 2, 3
        assert flags[0] == [
            int(ElevationType.HARMONIC),
            int(VelocityType.HARMONIC),
            0,
            0,
        ]
        assert flags[1] == [
            int(ElevationType.CONSTANT),
            int(VelocityType.CONSTANT),
            0,
            0,
        ]
        assert flags[2] == [0, 0, 0, 0]  # Default for missing index 2
        assert flags[3] == [int(ElevationType.NONE), int(VelocityType.FLATHER), 0, 0]

    def test_get_constant_values(self, sample_grid_path):
        """Test getting constant values from configurations."""
        boundary = TidalBoundary(grid_path=sample_grid_path)

        # Add configuration with constant values
        boundary.set_boundary_type(
            0,
            ElevationType.CONSTANT,
            VelocityType.CONSTANT,
            ethconst=1.0,
            vthconst=-100.0,
        )

        constants = boundary.get_constant_values()

        assert "ethconst" in constants
        assert "vthconst" in constants
        assert constants["ethconst"][0] == 1.0
        assert constants["vthconst"][0] == -100.0

    def test_create_bctides(self, sample_grid_path, sample_tidal_files, monkeypatch):
        """Test creating a Bctides object."""
        # Skip actual file operations in Bctides
        monkeypatch.setattr(Bctides, "_get_tidal_factors", lambda self: None)

        boundary = TidalBoundary(
            grid_path=sample_grid_path,
            constituents=["M2", "S2", "N2"],
            tidal_elevations=sample_tidal_files["elevations"],
            tidal_velocities=sample_tidal_files["velocities"],
        )

        boundary.set_boundary_type(0, ElevationType.HARMONIC, VelocityType.HARMONIC)

        # Set run parameters
        boundary.set_run_parameters(datetime(2023, 1, 1), 10.0)

        # Create Bctides object
        bctides = boundary.create_bctides()

        assert isinstance(bctides, Bctides)
        assert bctides._start_time == datetime(2023, 1, 1)
        assert bctides._rnday == 10.0
        assert bctides.tnames == ["M2", "S2", "N2"]


class TestFactoryFunctions:
    """Tests for the factory functions."""

    def test_create_tidal_boundary(self, sample_grid_path):
        """Test creating a tidal-only boundary."""
        boundary = create_tidal_boundary(
            grid_path=sample_grid_path, constituents=["M2", "S2", "N2"]
        )

        assert isinstance(boundary, TidalBoundary)
        assert boundary.constituents == ["M2", "S2", "N2"]

        # Check default configuration
        configs = boundary.boundary_configs
        assert 0 in configs
        assert configs[0].elev_type == ElevationType.HARMONIC
        assert configs[0].vel_type == VelocityType.HARMONIC

    def test_create_hybrid_boundary(self, sample_grid_path):
        """Test creating a hybrid boundary."""
        boundary = create_hybrid_boundary(
            grid_path=sample_grid_path, constituents=["M2", "S2", "N2"]
        )

        assert isinstance(boundary, TidalBoundary)

        # Check default configuration
        configs = boundary.boundary_configs
        assert 0 in configs
        assert configs[0].elev_type == ElevationType.HARMONICEXTERNAL
        assert configs[0].vel_type == VelocityType.HARMONICEXTERNAL

    def test_create_river_boundary(self, sample_grid_path):
        """Test creating a river boundary."""
        river_flow = -500.0
        boundary = create_river_boundary(
            grid_path=sample_grid_path, river_flow=river_flow
        )

        assert isinstance(boundary, TidalBoundary)

        # Check river configuration
        configs = boundary.boundary_configs
        assert 0 in configs
        assert configs[0].elev_type == ElevationType.NONE
        assert configs[0].vel_type == VelocityType.CONSTANT
        assert configs[0].vthconst == river_flow

    def test_create_nested_boundary(self, sample_grid_path):
        """Test creating a nested boundary."""
        boundary = create_nested_boundary(
            grid_path=sample_grid_path,
            with_tides=True,
            inflow_relax=0.9,
            outflow_relax=0.8,
            constituents=["M2", "S2", "N2"],
        )

        assert isinstance(boundary, TidalBoundary)

        # Check nested configuration
        configs = boundary.boundary_configs
        assert 0 in configs
        assert configs[0].elev_type == ElevationType.HARMONICEXTERNAL
        assert configs[0].vel_type == VelocityType.RELAXED
        assert configs[0].temp_type == TracerType.EXTERNAL
        assert configs[0].salt_type == TracerType.EXTERNAL
        assert configs[0].inflow_relax == 0.9
        assert configs[0].outflow_relax == 0.8



"""
Type handling for numpy types in Pydantic models.

This module provides utilities for handling numpy types in Pydantic models
without compromising type safety.
"""

from typing import Any

import numpy as np
from pydantic import field_validator

# Type aliases for type annotations
NumpyBool = bool  # We'll use validators instead of Union types
NumpyInt = int
NumpyFloat = float


# Function to convert numpy types to Python native types
def to_python_type(value: Any) -> Any:
    """Convert numpy types to Python native types."""
    if isinstance(value, np.bool_):
        return bool(value)
    elif isinstance(value, np.integer):
        return int(value)
    elif isinstance(value, np.floating):
        return float(value)
    elif isinstance(value, np.ndarray) and value.size == 1:
        # Handle single-element arrays
        return to_python_type(value.item())
    return value


# Reusable validators for Pydantic models
def numpy_bool_validator(v: Any) -> bool:
    """Validate and convert numpy boolean to Python boolean."""
    if isinstance(v, np.bool_):
        return bool(v)
    return v


def numpy_int_validator(v: Any) -> int:
    """Validate and convert numpy integer to Python integer."""
    if isinstance(v, np.integer):
        return int(v)
    return v


def numpy_float_validator(v: Any) -> float:
    """Validate and convert numpy float to Python float."""
    if isinstance(v, np.floating):
        return float(v)
    return v


# Create decorator functions for easy use in models
def validate_numpy_types(cls):
    """Class decorator to add numpy type validators to all fields."""
    for field_name in cls.model_fields:
        # Add appropriate validators based on field type
        field_info = cls.model_fields[field_name]
        if field_info.annotation in (bool, NumpyBool):
            cls.model_validators[f"_validate_{field_name}"] = field_validator(
                field_name
            )(numpy_bool_validator)
        elif field_info.annotation in (int, NumpyInt):
            cls.model_validators[f"_validate_{field_name}"] = field_validator(
                field_name
            )(numpy_int_validator)
        elif field_info.annotation in (float, NumpyFloat):
            cls.model_validators[f"_validate_{field_name}"] = field_validator(
                field_name
            )(numpy_float_validator)

    return cls



import os
import tempfile
from datetime import datetime
from pathlib import Path

from rompy_schism.bctides import Bctides
from rompy_schism.boundary_core import TidalBoundary  # Backward compatibility alias
from rompy_schism.boundary_core import (
    BoundaryConfig,
    ElevationType,
    VelocityType,
    create_tidal_boundary,
)


def validate_constituent_case_consistency(file_path):
    """Validate that the case of constituent names is consistent throughout the file."""
    with open(file_path, "r") as f:
        lines = f.readlines()

    # Remove comments and empty lines
    lines = [line.split("!")[0].strip() for line in lines]
    lines = [line for line in lines if line]

    # Initialize dictionary to track case usage
    constituents = {}

    # Process the file to find ntip, nbfr, and constituents
    i = 0

    # Parse ntip
    parts = lines[i].split()
    ntip = int(parts[0])
    i += 1

    # Skip tidal potential constituents
    if ntip > 0:
        for _ in range(ntip):
            constituent = lines[i].strip()
            if constituent not in constituents:
                constituents[constituent] = []
            constituents[constituent].append(i)
            i += 1

            # Skip parameters line
            i += 1

    # Parse nbfr
    nbfr = int(lines[i])
    i += 1

    # Parse frequency constituents
    for _ in range(nbfr):
        constituent = lines[i].strip()
        if constituent not in constituents:
            constituents[constituent] = []
        constituents[constituent].append(i)
        i += 1

        # Skip frequency line
        i += 1

    # Parse nope (number of open boundary segments)
    nope = int(lines[i])
    i += 1

    # Parse each boundary segment
    for j in range(nope):
        # Skip boundary flags line
        i += 1

        # Look for constituent lines in this segment (for elevation and velocity)
        while i < len(lines):
            line = lines[i]

            # Check if this line is a constituent name
            if line.upper() in [const.upper() for const in constituents.keys()]:
                # Find the matching constituent (case-insensitive)
                for const in constituents.keys():
                    if line.upper() == const.upper():
                        constituents[const].append(i)
                        break

            i += 1

            # Check if we've reached the end of the boundary segments
            if i < len(lines) and (
                lines[i].endswith("!ncbn") or lines[i].endswith("!nfluxf")
            ):
                break

    # For each constituent, check if the case is consistent
    inconsistent = []
    for const, positions in constituents.items():
        # Extract the actual strings at these positions
        instances = [lines[pos] for pos in positions]

        # Check if all instances have the same case
        if len(set(instances)) > 1:
            inconsistent.append(f"{const}: {instances}")

    return inconsistent


def test_tidal_boundary_constituent_consistency(
    grid2d, tidal_dataset, mock_tidal_data, monkeypatch
):
    """Test that constituent names in the bctides.in file have consistent case using real grid."""
    # Use the mock_tidal_data function for interpolation
    monkeypatch.setattr(Bctides, "_interpolate_tidal_data", mock_tidal_data)

    # Create boundary configs for tidal boundary
    configs = {}
    configs[0] = BoundaryConfig(
        id=0,
        elev_type=ElevationType.HARMONIC,
        vel_type=VelocityType.HARMONIC,
        temp_type=0,
        salt_type=0,
    )

    # Create a temporary file for output
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp_path = Path(tmp.name)

    try:
        # Get grid path from grid2d fixture
        grid_path = str(grid2d.hgrid.source)

        # Create a TidalBoundary instance with only the constituents in the test dataset (M2, S2, N2)
        boundary = TidalBoundary(
            grid_path=grid_path, boundary_configs=configs, tidal_data=tidal_dataset
        )

        # Set run parameters
        boundary.set_run_parameters(datetime(2023, 1, 1), 5.0)

        # Write the boundary file
        bctides_path = boundary.write_boundary_file(tmp_path)

        # Check for case consistency
        inconsistencies = validate_constituent_case_consistency(bctides_path)

        # Debug output
        with open(bctides_path, "r") as f:
            content = f.read()
            print(f"\nBCTIDES CONTENT:\n{content}\n")

        # There should be no inconsistencies
        assert (
            len(inconsistencies) == 0
        ), f"Inconsistent constituent cases: {inconsistencies}"

        # Check each constituent for case consistency
        for constituent in ["M2", "S2"]:
            upper_count = content.count(constituent)
            lower_count = content.count(constituent.lower())

            # Either all uppercase or all lowercase is acceptable, but mixing is not
            assert upper_count == 0 or lower_count == 0, f"Mixed case for {constituent}"

    finally:
        # Clean up
        if os.path.exists(tmp_path):
            os.unlink(tmp_path)


def test_create_tidal_boundary_wrapper(
    grid2d, tidal_dataset, mock_tidal_data, monkeypatch
):
    """Test the create_tidal_boundary wrapper function with real grid and data."""
    # Use the mock_tidal_data function for interpolation
    monkeypatch.setattr(Bctides, "_interpolate_tidal_data", mock_tidal_data)

    # Create a temporary file for output
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp_path = Path(tmp.name)

    try:
        # Get grid path from grid2d fixture
        grid_path = str(grid2d.hgrid.source)

        # Create the boundary with the wrapper function
        boundary = create_tidal_boundary(
            grid_path=grid_path,
            tidal_database=tidal_dataset.tidal_database,
            constituents=tidal_dataset.constituents,
            tidal_model=tidal_dataset.tidal_model,
        )

        # Set run parameters
        boundary.set_run_parameters(datetime(2023, 1, 1), 5.0)

        # Write the boundary file
        bctides_path = boundary.write_boundary_file(tmp_path)

        # Check for case consistency
        inconsistencies = validate_constituent_case_consistency(bctides_path)
        assert (
            len(inconsistencies) == 0
        ), f"Inconsistent constituent cases: {inconsistencies}"

    finally:
        # Clean up
        if os.path.exists(tmp_path):
            os.unlink(tmp_path)


def test_tidal_boundary_with_different_grids(
    request, grid2d, grid3d, tidal_dataset, mock_tidal_data, monkeypatch
):
    """Test tidal boundary with different grid types."""
    # Use the mock_tidal_data function for interpolation
    monkeypatch.setattr(Bctides, "_interpolate_tidal_data", mock_tidal_data)

    # Test with each grid type
    for grid_fixture in [grid2d, grid3d]:
        # Create a temporary file for output
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp_path = Path(tmp.name)

        try:
            # Get grid path
            grid_path = str(grid_fixture.hgrid.source)

            # Create the boundary
            boundary = create_tidal_boundary(
                grid_path=grid_path,
                tidal_database=tidal_dataset.tidal_database,
                constituents=tidal_dataset.constituents,
                tidal_model=tidal_dataset.tidal_model,
            )

            # Set run parameters
            boundary.set_run_parameters(datetime(2023, 1, 1), 5.0)

            # Write the boundary file
            bctides_path = boundary.write_boundary_file(tmp_path)

            # Check for case consistency
            inconsistencies = validate_constituent_case_consistency(bctides_path)
            assert (
                len(inconsistencies) == 0
            ), f"Inconsistent constituent cases: {inconsistencies}"

        finally:
            # Clean up
            if os.path.exists(tmp_path):
                os.unlink(tmp_path)


def test_case_consistency(grid2d, tidal_dataset, mock_tidal_data, monkeypatch):
    """Test that constituent names in the bctides.in file have consistent case."""
    # Use the mock_tidal_data function for interpolation
    monkeypatch.setattr(Bctides, "_interpolate_tidal_data", mock_tidal_data)

    # Create a temporary file for output
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp_path = Path(tmp.name)

    try:
        # Create a boundary
        grid_path = str(grid2d.hgrid.source)
        boundary = create_tidal_boundary(
            grid_path=grid_path,
            tidal_database=tidal_dataset.tidal_database,
            constituents=tidal_dataset.constituents,
            tidal_model=tidal_dataset.tidal_model,
        )

        # Write the boundary file
        boundary.set_run_parameters(datetime(2023, 1, 1), 5.0)
        boundary.write_boundary_file(tmp_path)

        # Check for case consistency
        inconsistencies = validate_constituent_case_consistency(tmp_path)
        assert (
            len(inconsistencies) == 0
        ), f"Found case inconsistencies: {inconsistencies}"

        # Check specifically for each constituent
        with open(tmp_path, "r") as f:
            content = f.read()

        for constituent in ["M2", "S2"]:
            # All occurrences should be same case (either all uppercase or all lowercase)
            upper_count = content.count(constituent)
            lower_count = content.count(constituent.lower())
            assert (
                upper_count == 0 or lower_count == 0
            ), f"Mixed case found for {constituent}"

    finally:
        # Clean up
        if os.path.exists(tmp_path):
            os.unlink(tmp_path)



# This file was auto generated from a schism namelist file on 2024-09-06.

from typing import Literal, Optional

from pydantic import Field, field_validator, model_validator

from rompy_schism.namelists.basemodel import NamelistBaseModel


class Proc(NamelistBaseModel):
    """Main processing parameters"""

    procname: str = Field("Isabel", description="Project Name")
    dimmode: Literal[2] = Field(
        2,
        description="Mode of run (ex: 1 = 1D, 2 = 2D) always 2D when coupled to SCHISM",
    )
    lstea: bool = Field(False, description="steady mode; under development")
    lqstea: bool = Field(
        False,
        description="Quasi-Steady Mode; In this case WWM-II is doing subiterations defined as DELTC/NQSITER unless QSCONVI is not reached",
    )
    lsphe: bool = Field(True, description="Spherical coordinates (lon/lat)")
    lnautin: bool = Field(
        True,
        description="Nautical convention for all inputs given in degrees (suggestion: T)",
    )
    lnautout: bool = Field(True, description="Output in Nautical convention")
    lmono_in: bool = Field(
        False,
        description="For prescribing monochromatic wave height Hmono as a boundary conditions; incident wave is defined as monochromatic wave height, which is Hmono = sqrt(2) * Hs",
    )
    lmono_out: bool = Field(False, description="Output wave heights in terms of Lmono")
    begtc: str = Field(
        "20030908.000000",
        description="Time for start the simulation, ex:yyyymmdd. hhmmss",
    )
    deltc: int = Field(
        600, description="Time step (MUST match dt*nstep_wwm in SCHISM!)"
    )
    unitc: Literal["SEC"] = Field("SEC", description="Unity of time step")
    endtc: str = Field(
        "20031008.000000",
        description="Time for stop the simulation, ex:yyyymmdd. hhmmss",
    )
    dmin: float = Field(
        0.01, description="Minimum water depth. This must be the same as h0 in SCHISM"
    )


class Coupl(NamelistBaseModel):
    """Coupling parameters"""

    lcpl: bool = Field(
        True,
        description="Couple with current model ... main switch - keep it on for SCHISM-WWM",
    )
    radflag: Literal["LON", "VOR"] = Field(
        "LON", description="LON: Longuet-Higgin; VOR: vortex formulation"
    )
    letot: bool = Field(
        False, description="Option to compute the wave induced radiation stress"
    )
    nlvt: int = Field(10, description="Number of vertical Layers; not used with SCHISM")
    dtcoup: float = Field(
        600.0, description="Couple time step - not used when coupled to SCHISM"
    )


class Grid(NamelistBaseModel):
    """Grid parameters"""

    lcird: bool = Field(True, description="Full circle in directional space")
    lstag: bool = Field(
        False,
        description="Stagger directional bins with a half Dtheta; may use T only for regular grid to avoid char. line aligning with grid line",
    )
    mindir: float = Field(
        0.0,
        description="Minimum direction for simulation (unit: degrees; nautical convention; 0: from N; 90: from E); not used if LCIRD = .T.",
    )
    maxdir: float = Field(
        360.0,
        description="Maximum direction for simulation (unit: degrees); may be < MINDIR; not used if LCIRD = .T.",
    )
    mdc: int = Field(36, description="Number of directional bins")
    frlow: float = Field(
        0.04,
        description="Low frequency limit of the discrete wave period (Hz; 1/period)",
    )
    frhigh: float = Field(
        1.0, description="High frequency limit of the discrete wave period."
    )
    msc: int = Field(36, description="Number of frequency bins")
    filegrid: str = Field(
        "hgrid_WWM.gr3",
        description="Name of the grid file. hgridi_WWM.gr3 if IGRIDTYPE = 3 (SCHISM)",
    )
    igridtype: Literal[3] = Field(3, description="Gridtype used")
    lslop: bool = Field(False, description="Bottom Slope limiter (default=F)")
    slmax: float = Field(0.2, description="Max Slope")
    lvar1d: bool = Field(
        False, description="For 1d-mode if variable dx is used; not used with SCHISM"
    )
    loptsig: bool = Field(
        False,
        description="Use optimal distributions of freq. in spectral space ... fi+1 = fi * 1.1. Take care what you high freq. limit is!",
    )


class Init(NamelistBaseModel):
    """Initialization parameters"""

    lhotr: bool = Field(
        False, description="Use hotstart file (see &HOTFILE section for input file)"
    )
    linid: bool = Field(
        True,
        description="Initial condition; F for default; use T if using WW3 as i.c. etc",
    )
    initstyle: Literal[1, 2] = Field(
        2,
        description="1 - Parametric Jonswap, 2 - Read from Global NETCDF files, work only if IBOUNDFORMAT=3",
    )


class Bouc(NamelistBaseModel):
    """Boundary condition parameters"""

    lbcse: bool = Field(True, description="The wave boundary data is time dependent")
    lbinter: bool = Field(
        True,
        description="Do interpolation in time if LBCSE=T (not available for quasi-steady mode within the subtime steps)",
    )
    lbcwa: bool = Field(True, description="Parametric Wave Spectra")
    lbcsp: bool = Field(
        False,
        description="Specify (non-parametric) wave spectra, specified in 'FILEWAVE' below",
    )
    linhom: bool = Field(True, description="Non-uniform wave b.c. in space")
    lbsp1d: bool = Field(
        False,
        description="1D (freq. space only) format for FILEWAVE if LBCSP=T and LINHOM=F",
    )
    lbsp2d: bool = Field(
        False, description="2D format for FILEWAVE if LBCSP=T and LINHOM=F"
    )
    begtc: str = Field(
        "20030908.000000", description="Begin time of the wave boundary file (FILEWAVE)"
    )
    deltc: int = Field(1, description="Time step in FILEWAVE")
    unitc: Literal["HR", "MIN", "SEC"] = Field(
        "HR", description="Unit can be HR, MIN, SEC"
    )
    endtc: str = Field("20031008.000000", description="End time")
    filebound: str = Field(
        "wwmbnd.gr3",
        description="Boundary file defining boundary conditions and Neumann nodes",
    )
    iboundformat: Literal[1, 3, 6] = Field(
        3,
        description="1 ~ WWM, 3 ~ WW3 (2D spectra in netcdf format only - LBCWA=T), 6 ~ WW3 2D spectra in netcdf format with LBCSP=T (prescribed spectra)",
    )
    filewave: str = Field(
        "bndfiles.dat", description="Boundary file defining boundary input from WW3"
    )
    lindsprdeg: bool = Field(
        False,
        description="If 1-d wave spectra are read this flag defines whether the input for the directional spreading is in degrees (true) or exponent (false)",
    )
    lparmdir: bool = Field(
        False,
        description="If LPARMDIR is true then directional spreading is read from WBDS and must be in exponential format at this time, only valid for 1d Spectra",
    )
    wbhs: float = Field(2.0, description="Hs at the boundary for parametric spectra")
    wbss: Literal[1, -1, 2, -2, 3, -3, 4] = Field(
        2, description="Spectral shape parameter"
    )
    wbtp: float = Field(
        8.0,
        description="Tp at the boundary (sec); mean or peak depending on the sign of WBSS",
    )
    wbdm: float = Field(90.0, description="Avg. Wave Direction at the boundary")
    wbdsms: Literal[1, 2] = Field(
        1, description="Directional spreading value in degrees (1) or as exponent (2)"
    )
    wbds: float = Field(
        10.0, description="Directional spreading at the boundary (degrees/exponent)"
    )
    wbgauss: float = Field(
        0.1, description="factor for gaussian distribution if WBSS=1"
    )
    wbpken: float = Field(
        3.3, description="Peak enhancement factor for Jonswap Spectra if WBSS=2"
    )

    @field_validator("wbss")
    def validate_wbss(cls, v):
        if v not in [1, -1, 2, -2, 3, -3, 4]:
            raise ValueError("WBSS must be 1, -1, 2, -2, 3, -3, or 4")
        return v


class Engs(NamelistBaseModel):
    """Source terms parameters"""

    mesnl: Literal[0, 1] = Field(
        1,
        description="Nonlinear Interaction NL4 , 1 ~ on, 0 ~ off (Discrete Interaction approx.)",
    )
    mesin: Literal[0, 1, 2, 3, 4, 5] = Field(1, description="Wind input formulation")
    ifric: Literal[1, 4] = Field(
        1, description="Formulation for atmospheric boundary layer"
    )
    mesbf: Literal[1, 2, 3] = Field(1, description="Bottom friction formulation")
    fricc: float = Field(0.067, description="Bottom friction coefficient or roughness")
    mesbr: Literal[0, 1] = Field(
        1, description="Shallow water wave breaking; 0: off; 1: on"
    )
    ibreak: Literal[1, 2, 3, 4, 5, 6] = Field(
        1, description="Wave breaking formulation"
    )
    icrit: Literal[1, 2, 3, 4, 5, 6] = Field(1, description="Wave breaking criterion")
    brcr: float = Field(0.78, description="Wave breaking parameter")
    a_brcr: float = Field(0.76, description="Wave breaking parameter")
    b_brcr: float = Field(0.29, description="Wave breaking parameter")
    min_brcr: float = Field(0.25, description="Wave breaking parameter")
    max_brcr: float = Field(0.8, description="Wave breaking parameter")
    a_biph: float = Field(0.2, description="Biphase coefficient")
    br_coef_method: Literal[1, 2] = Field(
        1, description="Method for the breaking coefficient"
    )
    b_alp: float = Field(0.5, description="breaking coefficient")
    zprof_break: Literal[1, 2, 3, 4, 5, 6] = Field(
        2, description="Vertical distribution function of wave breaking source term"
    )
    bc_break: Literal[0, 1] = Field(
        1, description="Apply depth-limited breaking at the boundaries"
    )
    iroller: Literal[0, 1] = Field(0, description="Wave roller model")
    alprol: float = Field(
        0.85, description="Alpha coefficient for the wave roller model"
    )
    meveg: Literal[0, 1] = Field(0, description="Vegetation on/off")
    lmaxetot: bool = Field(
        True, description="Limit shallow water wave height by wave breaking limiter"
    )
    mesds: Literal[1, 2] = Field(
        1, description="Formulation for the whitecapping source function"
    )
    mestr: Literal[0, 1, 2] = Field(
        1, description="Formulation for the triad 3 wave interactions"
    )
    trico: float = Field(0.1, description="proportionality const. (\\alpha_EB)")
    trira: float = Field(
        2.5, description="ratio of max. freq. considered in triads over mean freq."
    )
    triurs: float = Field(0.1, description="critical Ursell number")


class Nums(NamelistBaseModel):
    """Numerical parameters"""

    icomp: Literal[0, 1, 2, 3] = Field(
        3,
        description="Controls splitting method and implicit/explicit schemes for spectral advection",
    )
    amethod: Literal[0, 1, 2, 3, 4, 5, 6, 7] = Field(
        7, description="Controls different Methods in geographical space"
    )
    smethod: Literal[0, 1, 2, 3, 4, 6] = Field(
        1, description="Controls the way the source terms are integrated"
    )
    dmethod: Literal[0, 1, 2, 3, 4] = Field(
        2, description="Controls the numerical method in directional space"
    )
    rtheta: float = Field(0.5, description="Weighing factor for DMETHOD = 1")
    litersplit: bool = Field(
        False, description="T: double Strang split; F: simple split (more efficient)"
    )
    lfilterth: bool = Field(
        False,
        description="Use a CFL filter to limit the advection vel. In directional space",
    )
    maxcflth: float = Field(
        1.0, description="Max Cfl in Theta space; used only if LFILTERTH=T"
    )
    fmethod: Literal[0, 1] = Field(
        1, description="Controls the numerical method used in freq. space"
    )
    lfiltersig: bool = Field(
        False, description="Limit the advection velocity in freq. space"
    )
    maxcflsig: float = Field(
        1.0, description="Max Cfl in freq. space; used only if LFILTERSIG=T"
    )
    llimt: bool = Field(True, description="Switch on/off Action limiter")
    melim: Literal[1, 2, 3] = Field(1, description="Formulation for the action limiter")
    limfak: float = Field(
        0.1, description="Proportionality coefficient for the action limiter"
    )
    ldifr: bool = Field(
        False, description="Use phase decoupled diffraction approximation"
    )
    idiffr: Literal[1, 2, 3] = Field(
        1, description="Extended WAE accounting for higher order effects"
    )
    lconv: bool = Field(
        False, description="Estimate convergence criterian and write disk"
    )
    lcfl: bool = Field(False, description="Write out CFL numbers")
    nqsiter: int = Field(
        1, description="# of quasi-steady (Q-S) sub-divisions within each WWM time step"
    )
    qsconv1: float = Field(
        0.98,
        description="Number of grid points [%/100] that have to fulfill abs. wave height criteria EPSH1",
    )
    qsconv2: float = Field(
        0.98,
        description="Number of grid points [%/100] that have to fulfill rel. wave height criteria EPSH2",
    )
    qsconv3: float = Field(
        0.98,
        description="Number of grid points [%/100] that have to fulfill sum. rel. wave action criteria EPSH3",
    )
    qsconv4: float = Field(
        0.98,
        description="Number of grid points [%/100] that have to fulfill rel. avg. wave steepness criteria EPSH4",
    )
    qsconv5: float = Field(
        0.98,
        description="Number of grid points [%/100] that have to fulfill avg. rel. wave period criteria EPSH5",
    )
    lexpimp: bool = Field(
        False,
        description="Use implicit schemes for freq. lower than given below by FREQEXP",
    )
    freqexp: float = Field(0.1, description="Minimum frequency for explicit schemes")
    epsh1: float = Field(0.01, description="Convergence criteria for rel. wave height")
    epsh2: float = Field(0.01, description="Convergence criteria for abs. wave height")
    epsh3: float = Field(
        0.01, description="Convergence criteria for the rel. sum of wave action"
    )
    epsh4: float = Field(
        0.01,
        description="Convergence criteria for the rel. avg. wave steepness criteria",
    )
    epsh5: float = Field(
        0.01, description="Convergence criteria for the rel. avg. waveperiod"
    )
    lvector: bool = Field(
        False,
        description="Use optmized propagation routines for large high performance computers",
    )
    ivector: Literal[1, 2, 3, 4, 5, 6] = Field(
        2, description="Different flavours of communications"
    )
    ladvtest: bool = Field(False, description="for testing the advection schemes")
    lchkconv: bool = Field(
        False, description="needs to set to .true. for quasi-steady mode"
    )
    dtmin_dyn: float = Field(
        1.0, description="min. time step (sec?) for dynamic integration"
    )
    ndyniter: int = Field(100, description="max. iteration for dyn. scheme")
    dtmin_sin: float = Field(
        1.0, description="min. time steps for the full fractional step method"
    )
    dtmin_snl4: float = Field(1.0, description="min. time steps for SNL4")
    dtmin_sds: float = Field(1.0, description="min. time steps for SDS")
    dtmin_snl3: float = Field(1.0, description="min. time steps for SNL3")
    dtmin_sbr: float = Field(0.10, description="min. time steps for SBR")
    dtmin_sbf: float = Field(1.0, description="min. time steps for SBF")
    ndyniter_sin: int = Field(10, description="max. iterations for SIN")
    ndyniter_snl4: int = Field(10, description="max. iterations for SNL4")
    ndyniter_sds: int = Field(10, description="max. iterations for SDS")
    ndyniter_sbr: int = Field(10, description="max. iterations for SBR")
    ndyniter_snl3: int = Field(10, description="max. iterations for SNL3")
    ndyniter_sbf: int = Field(10, description="max. iterations for SBF")
    lsoubound: bool = Field(False, description="Do source terms on boundary")
    wae_solverthr: float = Field(
        1e-6, description="Threshold for the Block-Jacobi or Block-Gauss-Seider solver"
    )
    maxiter: int = Field(1000, description="Max. number of iterations")
    pmin: float = Field(1.0, description="Max. percentage of non-converged grid points")
    lnaninfchk: bool = Field(False, description="Check for NaN and INF")
    lzeta_setup: bool = Field(False, description="Compute wave setup")
    zeta_meth: Literal[0] = Field(0, description="Method for wave setup")
    lsourceswam: bool = Field(
        False, description="Use ECMWF WAM formualtion for deep water physics"
    )
    block_gauss_seidel: bool = Field(
        True, description="Use the Gauss Seidel on each computer block"
    )
    lnonl: bool = Field(
        False,
        description="Solve the nonlinear system using simpler algorithm (Patankar)",
    )
    aspar_local_level: Literal[0, 1, 2, 3, 4, 5] = Field(
        0, description="Aspar locality level"
    )
    l_solver_norm: bool = Field(
        False,
        description="Compute solver norm ||A*x-b|| as termination check of jacobi-Gauss-Seidel solver",
    )
    laccel: bool = Field(False, description="Acceleration flag")


class History(NamelistBaseModel):
    """History output parameters"""

    begtc: str = Field(
        "20030908.000000", description="Start output time, yyyymmdd. hhmmss"
    )
    deltc: int = Field(1, description="Time step for output")
    unitc: Literal["SEC"] = Field("SEC", description="Unit")
    endtc: str = Field(
        "20031008.000000", description="Stop time output, yyyymmdd. hhmmss"
    )
    definetc: int = Field(86400, description="Time scoop (sec) for history files")
    outstyle: Literal["NO", "NC", "XFN", "SHP"] = Field(
        "NO", description="output option"
    )
    multipleout: Literal[0, 1] = Field(0, description="Output file format")
    use_single_out: bool = Field(
        True, description="Use single precision in the output of model variables"
    )
    paramwrite: bool = Field(
        True,
        description="Write the physical parametrization and chosen numerical method in the netcdf file",
    )
    gridwrite: bool = Field(
        True, description="Write the grid in the netcdf history file"
    )
    printmma: bool = Field(
        False,
        description="Print minimum, maximum and average value of statistics during runtime",
    )
    fileout: str = Field("wwm_hist.dat", description="Output file name")
    hs: bool = Field(False, description="Output significant wave height")
    tm01: bool = Field(False, description="Output mean period")
    tm02: bool = Field(False, description="Output zero-crossing mean period")
    klm: bool = Field(False, description="Output mean wave number")
    wlm: bool = Field(False, description="Output mean wave length")
    etotc: bool = Field(False, description="Output variable ETOTC")
    etots: bool = Field(False, description="Output variable ETOTS")
    dm: bool = Field(False, description="Output mean wave direction")
    dspr: bool = Field(False, description="Output directional spreading")
    tppd: bool = Field(False, description="Output direction of the peak")
    tpp: bool = Field(False, description="Output peak period")
    cpp: bool = Field(False, description="Output peak phase vel.")
    wnpp: bool = Field(False, description="Output peak wave number")
    cgpp: bool = Field(False, description="Output peak group speed")
    kpp: bool = Field(False, description="Output peak wave number")
    lpp: bool = Field(False, description="Output peak wave length")
    peakd: bool = Field(False, description="Output peak direction")
    peakdspr: bool = Field(False, description="Output peak directional spreading")
    dpeak: bool = Field(False, description="Output peak direction")
    ubot: bool = Field(False, description="Output bottom exc. vel.")
    orbital: bool = Field(False, description="Output bottom orbital vel.")
    botexper: bool = Field(False, description="Output bottom exc.")
    tmbot: bool = Field(False, description="Output bottom period")
    ursell: bool = Field(False, description="Output Ursell number")
    ufric: bool = Field(False, description="Output air friction velocity")
    z0: bool = Field(False, description="Output air roughness length")
    alpha_ch: bool = Field(False, description="Output Charnoch coefficient for air")
    windx: bool = Field(False, description="Output Wind in X direction")
    windy: bool = Field(False, description="Output Wind in Y direction")
    cd: bool = Field(False, description="Output Drag coefficient")
    currtx: bool = Field(False, description="Output current in X direction")
    currty: bool = Field(False, description="Output current in Y direction")
    watlev: bool = Field(False, description="Output water level")
    watlevold: bool = Field(
        False, description="Output water level at previous time step"
    )
    depdt: bool = Field(False, description="Output change of water level in time")
    dep: bool = Field(False, description="Output depth")
    tauw: bool = Field(False, description="Output surface stress from the wave")
    tauhf: bool = Field(False, description="Output high frequency surface stress")
    tautot: bool = Field(False, description="Output total surface stress")
    stokessurfx: bool = Field(
        False, description="Output Surface Stokes drift in X direction"
    )
    stokessurfy: bool = Field(
        False, description="Output Surface Stokes drift in Y direction"
    )
    stokesbarox: bool = Field(
        False, description="Output Barotropic Stokes drift in X direction"
    )
    stokesbaroy: bool = Field(
        False, description="Output Barotropic Stokes drift in Y direction"
    )
    rsxx: bool = Field(False, description="Output RSXX potential of LH")
    rsxy: bool = Field(False, description="Output RSXY potential of LH")
    rsyy: bool = Field(False, description="Output RSYY potential of LH")
    cfl1: bool = Field(False, description="Output CFL number 1")
    cfl2: bool = Field(False, description="Output CFL number 2")
    cfl3: bool = Field(False, description="Output CFL number 3")


class Station(NamelistBaseModel):
    """Station output parameters"""

    begtc: str = Field(
        "20030908.000000", description="Start simulation time, yyyymmdd. hhmmss"
    )
    deltc: int = Field(600, description="Time step for output")
    unitc: Literal["SEC"] = Field("SEC", description="Unit")
    endtc: str = Field(
        "20031008.000000", description="Stop time simulation, yyyymmdd. hhmmss"
    )
    definetc: int = Field(86400, description="Time for definition of station files")
    outstyle: Literal["NO", "STE", "NC"] = Field("NO", description="output option")
    multipleout: Literal[0, 1] = Field(0, description="Output file format")
    use_single_out: bool = Field(
        True, description="Use single precision in the output of model variables"
    )
    paramwrite: bool = Field(
        True,
        description="Write the physical parametrization and chosen numerical method in the netcdf file",
    )
    fileout: str = Field("wwm_sta.dat", description="Output file name")
    loutiter: bool = Field(False, description="Output iteration flag")
    iouts: int = Field(15, description="Number of output stations")
    nouts: list[str] = Field(
        [
            "P-1",
            "P-2",
            "P-3",
            "P-4",
            "P-5",
            "P-6",
            "P-7",
            "P-8",
            "P-9",
            "P-10",
            "P-11",
            "P-12",
            "P-13",
            "P-14",
            "P-15",
        ],
        description="Names of output stations",
    )
    xouts: list[float] = Field(
        [
            -76.046,
            -76.778,
            -75.81,
            -75.72,
            -74.842,
            -74.703,
            -75.33,
            -72.631,
            -74.835,
            -69.248,
            -72.6,
        ],
        description="X coordinates of output stations",
    )
    youts: list[float] = Field(
        [
            39.152,
            38.556,
            38.033,
            37.551,
            36.974,
            37.204,
            37.023,
            36.915,
            36.611,
            38.461,
            35.75,
            34.561,
            31.862,
            40.503,
            39.584,
        ],
        description="Y coordinates of output stations",
    )
    cutoff: list[float] = Field(
        [0.44] * 15, description="cutoff freq (Hz) for each station"
    )
    lsp1d: bool = Field(True, description="1D spectral station output")
    lsp2d: bool = Field(False, description="2D spectral station output")
    lsigmax: bool = Field(True, description="Adjust the cut-freq. for the output")
    ac: bool = Field(False, description="Output spectrum")
    wk: bool = Field(False, description="Output variable WK")
    acout_1d: bool = Field(False, description="Output variable ACOUT_1D")
    acout_2d: bool = Field(False, description="Output variable ACOUT_2D")
    hs: bool = Field(False, description="Output significant wave height")
    tm01: bool = Field(False, description="Output mean period")
    tm02: bool = Field(False, description="Output zero-crossing mean period")
    klm: bool = Field(False, description="Output mean wave number")
    wlm: bool = Field(False, description="Output mean wave length")
    etotc: bool = Field(False, description="Output Variable ETOTC")
    etots: bool = Field(False, description="Output Variable ETOTS")
    dm: bool = Field(False, description="Output mean wave direction")
    dspr: bool = Field(False, description="Output directional spreading")
    tppd: bool = Field(False, description="Output Discrete Peak Period")
    tpp: bool = Field(False, description="Output Peak Period")
    cpp: bool = Field(False, description="Output CPP")
    wnpp: bool = Field(False, description="Output peak wave number")
    cgpp: bool = Field(False, description="Output peak group speed")
    kpp: bool = Field(False, description="Output peak wave number")
    lpp: bool = Field(False, description="Output peak")
    peakd: bool = Field(False, description="Output peak direction")
    peakdspr: bool = Field(False, description="Output peak directional spreading")
    dpeak: bool = Field(False, description="Output DPEAK")
    ubot: bool = Field(False, description="Output UBOT")
    orbital: bool = Field(False, description="Output ORBITAL")
    botexper: bool = Field(False, description="Output BOTEXPER")
    tmbot: bool = Field(False, description="Output TMBOT")
    ursell: bool = Field(False, description="Output Ursell number")
    ufric: bool = Field(False, description="Output air friction velocity")
    z0: bool = Field(False, description="Output air roughness length")
    alpha_ch: bool = Field(False, description="Output Charnoch coefficient for air")
    windx: bool = Field(False, description="Output Wind in X direction")
    windy: bool = Field(False, description="Output Wind in Y direction")
    cd: bool = Field(False, description="Output Drag coefficient")
    currtx: bool = Field(False, description="Output current in X direction")
    currty: bool = Field(False, description="Output current in Y direction")
    watlev: bool = Field(False, description="Output water level")
    watlevold: bool = Field(
        False, description="Output water level at previous time step"
    )
    depdt: bool = Field(False, description="Output change of water level in time")
    dep: bool = Field(False, description="Output depth")
    tauw: bool = Field(False, description="Output surface stress from the wave")
    tauhf: bool = Field(False, description="Output high frequency surface stress")
    tautot: bool = Field(False, description="Output total surface stress")
    stokessurfx: bool = Field(
        False, description="Output Surface Stokes drift in X direction"
    )
    stokessurfy: bool = Field(
        False, description="Output Surface Stokes drift in Y direction"
    )
    stokesbarox: bool = Field(
        False, description="Output Barotropic Stokes drift in X direction"
    )
    stokesbaroy: bool = Field(
        False, description="Output Barotropic Stokes drift in Y direction"
    )
    rsxx: bool = Field(False, description="Output RSXX potential of LH")
    rsxy: bool = Field(False, description="Output RSXY potential of LH")
    rsyy: bool = Field(False, description="Output RSYY potential of LH")
    cfl1: bool = Field(False, description="Output CFL number 1")
    cfl2: bool = Field(False, description="Output CFL number 2")
    cfl3: bool = Field(False, description="Output CFL number 3")


class Hotfile(NamelistBaseModel):
    """Configuration for WWM hotfile input and output."""

    lhotf: bool = Field(True, description="Write hotfile")
    filehot_out: str = Field("hotfile_out_WWM.nc", description="Name of output hotfile")
    begtc: str = Field(
        "20030908.000000", description="Starting time of hotfile writing"
    )
    deltc: float = Field(86400.0, description="Time between hotfile writes")
    unitc: Literal["SEC"] = Field("SEC", description="Unit used for time")
    endtc: str = Field("20031008.000000", description="Ending time of hotfile writing")
    lcyclehot: bool = Field(True, description="Cyclic hotfile writing for netcdf")
    hotstyle_out: Literal[1, 2] = Field(
        2, description="Output hotfile style (1: binary, 2: netcdf)"
    )
    multipleout: Literal[0, 1] = Field(0, description="Single or multiple output files")
    filehot_in: str = Field("hotfile_in_WWM.nc", description="Input hotfile name")
    hotstyle_in: Literal[1, 2] = Field(
        2, description="Input hotfile style (1: binary, 2: netcdf)"
    )
    ihotpos_in: Literal[1, 2] = Field(
        1, description="Position in input hotfile for reading"
    )
    multiplein: Literal[0, 1] = Field(0, description="Single or multiple input files")

    @field_validator("begtc", "endtc")
    def validate_datetime(cls, v):
        try:
            datetime.strptime(v, "%Y%m%d.%H%M%S")
        except ValueError:
            raise ValueError("Incorrect date format, should be YYYYMMDD.HHMMSS")
        return v

    @model_validator(mode="after")
    def check_time_range(self):
        start = datetime.strptime(self.begtc, "%Y%m%d.%H%M%S")
        end = datetime.strptime(self.endtc, "%Y%m%d.%H%M%S")
        if start >= end:
            raise ValueError("BEGTC must be earlier than ENDTC")
        return self


class PetscOptions(NamelistBaseModel):
    """Configuration for PETSc linear solver options."""

    ksptype: Literal["LGMRES", "GMRES", "DGMRES", "PGMRES", "KSPBCGSL"] = Field(
        "LGMRES", description="PETSc solver type"
    )
    rtol: float = Field(1e-20, description="Relative convergence tolerance")
    abstol: float = Field(1e-20, description="Absolute convergence tolerance")
    dtol: float = Field(10000.0, description="Divergence tolerance")
    maxits: int = Field(1000, description="Maximum number of iterations")
    initialguessnonzero: bool = Field(False, description="Initial guess is nonzero")
    gmrespreallocate: bool = Field(True, description="Preallocate GMRES work vectors")
    pctype: Literal["SOR", "ASM", "HYPRE", "SPAI", "NONE"] = Field(
        "SOR", description="PETSc preconditioner type"
    )

    @field_validator("rtol", "abstol")
    def validate_tolerance(cls, v):
        if v <= 0:
            raise ValueError("Tolerance must be positive")
        return v

    @field_validator("maxits")
    def validate_maxits(cls, v):
        if v <= 0:
            raise ValueError("Maximum iterations must be positive")
        return v


class Nesting(NamelistBaseModel):
    """Configuration for WWM nesting options."""

    list_begtc: str = Field("", description="List of beginning times for nesting")
    list_deltc: float = Field(0.0, description="List of time steps for nesting")
    list_unitc: str = Field("", description="List of time units for nesting")
    list_endtc: str = Field("", description="List of ending times for nesting")
    list_igridtype: int = Field(0, description="List of grid types for nesting")
    list_filegrid: str = Field("", description="List of grid files for nesting")
    list_filebound: str = Field("", description="List of boundary files for nesting")
    list_prefix: str = Field("", description="List of prefixes for nesting")


class Wwminput_WW3(NamelistBaseModel):
    """Main configuration for WWM (Wind Wave Model) based on Hurricane Isabel for Chesapeake Bay test."""

    hotfile: Optional[Hotfile] = Field(default=None)
    petscoptions: Optional[PetscOptions] = Field(default=None)
    nesting: Optional[Nesting] = Field(default=None)



# create pydantic model from the following using the Fields notation with defaults as provided here and descriptions taken from the comments
from pydantic import Field

from rompy.core.types import RompyBaseModel


class MARCO(RompyBaseModel):
    """
    Number of subcycles in ICM kinetics
    """

    nsub: int = Field(1, description="Number of subcycles in ICM kinetics")
    iKe: int = Field(
        0,
        description="Options of formulations for computing light attenuation coefficients",
    )
    Ke0: float = Field(
        0.26, description="Background light extinction coefficient (1/m)"
    )
    KeC: float = Field(0.017, description="Light attenuation due to chlorophyll")
    KeS: float = Field(0.07, description="Light attenuation due to TSS")
    KeSalt: float = Field(
        -0.02, description="Light attenuation due to CDOM (related to salinity)"
    )
    tss2c: float = Field(6.0, description="TSS to carbon ratio")
    iLight: int = Field(0, description="Options of computing light limitation factor")
    alpha: list[float] = Field(
        [8.0, 8.0, 8.0],
        description="Initial slope of P-I curve (g[C]*m2/g[Chl]/E), (iLight=0)",
    )
    iPR: int = Field(1, description="Options of phytoplankton' predation term")
    PRR: list[float] = Field(
        [0.1, 0.2, 0.05],
        description="Predation rate by higher trophic level (day-1, or day-1.g-1.m3)",
    )
    wqc0: list[float] = Field(
        [
            1.0,
            0.5,
            0.05,
            1.0,
            0.5,
            0.5,
            0.15,
            0.15,
            0.05,
            0.01,
            0.05,
            0.005,
            0.005,
            0.01,
            0.05,
            0.0,
            12.0,
        ],
        description="ICM init. value (wqc0), settling velocity (WSP), net settling velocity (WSPn) (m.day-1)",
    )
    WSP: list[float] = Field(
        [
            0.3,
            0.10,
            0.0,
            0.25,
            0.25,
            0.0,
            0.25,
            0.25,
            0.0,
            0.0,
            0.0,
            0.25,
            0.25,
            0.0,
            1.0,
            0.0,
            0.0,
        ],
        description="ICM init. value (wqc0), settling velocity (WSP), net settling velocity (WSPn) (m.day-1)",
    )
    WSPn: list[float] = Field(
        [
            0.3,
            0.10,
            0.0,
            0.25,
            0.25,
            0.0,
            0.25,
            0.25,
            0.0,
            0.0,
            0.0,
            0.25,
            0.25,
            0.0,
            1.0,
            0.0,
            0.0,
        ],
        description="ICM init. value (wqc0), settling velocity (WSP), net settling velocity (WSPn) (m.day-1)",
    )
    iSilica: int = Field(0, description="Silica model (0: OFF; 1: ON)")
    iZB: int = Field(
        0, description="Zooplankton (iZB=1: use zooplankton dynamics; iZB=0: don't use)"
    )
    iPh: int = Field(0, description="PH model (0: OFF; 1: ON)")
    iCBP: int = Field(0, description="ChesBay Program Model (0: OFF; 1: ON)")
    isav_icm: int = Field(0, description="Submerged Aquatic Vegetation switch")
    iveg_icm: int = Field(0, description="Intertidal vegetation switch")
    iSed: int = Field(
        1, description="Sediment module switch. iSed=1: Use sediment flux model"
    )
    iBA: int = Field(0, description="Benthic Algae")
    iRad: int = Field(0, description="Solar radiation")
    isflux: int = Field(0, description="Atmospheric fluxes")
    ibflux: int = Field(0, description="Bottom fluxes")
    iout_icm: int = Field(0, description="ICM station outputs")
    nspool_icm: int = Field(24, description="ICM station outputs")
    iLimit: int = Field(
        0, description="Options of nutrient limitation on phytoplankton growth"
    )
    idry_icm: int = Field(
        0,
        description="idry_icm=1: turn on shallow kinetic biochemical process; idry_icm=0: jump dry elems, keep last wet value; jump won't happen if iveg_icm is turn on and this elem is veg",
    )


#


class CORE(RompyBaseModel):
    """
    ICM parameters for water column; PB: phytoplankton
    CBP module parameters are included: (SRPOC,SRPON,SRPOP,PIP)
    """

    GPM: list[float] = Field([2.5, 2.8, 3.5], description="PB growth rates (day-1)")
    TGP: list[float] = Field(
        [15.0, 22.0, 27.0], description="optimal temp. for PB growth (oC)"
    )
    KTGP: list[float] = Field(
        [0.005, 0.004, 0.003, 0.008, 0.006, 0.004],
        description="temp. dependence for PB growth; dim(PB=1:3,1:2) (oC-2)",
    )
    MTR: list[float] = Field(
        [0.0, 0.0, 0.0], description="PB photorespiration coefficient (0<MTR<1)"
    )
    MTB: list[float] = Field(
        [0.01, 0.02, 0.03], description="PB metabolism rates (day-1)"
    )
    TMT: list[float] = Field(
        [20.0, 20.0, 20.0], description="reference temp. for PB metabolism (oC)"
    )
    KTMT: list[float] = Field(
        [0.0322, 0.0322, 0.0322],
        description="temp. dependence for PB metabolism (oC-1)",
    )
    FCP: list[float] = Field(
        [0.35, 0.30, 0.20, 0.55, 0.50, 0.50, 0.10, 0.20, 0.30, 0.0, 0.0, 0.0],
        description="fractions of PB carbon into (RPOC,LPOC,DOC,SRPOC); dim(PB=1:3,3)",
    )
    FNP: list[float] = Field(
        [
            0.35,
            0.35,
            0.35,
            0.50,
            0.50,
            0.50,
            0.10,
            0.10,
            0.10,
            0.05,
            0.05,
            0.05,
            0.0,
            0.0,
            0.0,
        ],
        description="fractions of PB nitrogen into (RPON,LPON,DON,NH4,SRPON)",
    )
    FPP: list[float] = Field(
        [
            0.10,
            0.10,
            0.10,
            0.20,
            0.20,
            0.20,
            0.50,
            0.50,
            0.50,
            0.20,
            0.20,
            0.20,
            0.0,
            0.0,
            0.0,
        ],
        description="fractions of PB Phosphorus into (RPOP,LPOP,DOP,PO4,SRPOP)",
    )
    FCM: list[float] = Field(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.1, 0.1, 0.0, 0.0, 0.0],
        description="fractions of PB metabolism carbon into (RPOC,LPOC,DOC,SRPOC); dim(PB=1:3,4)",
    )
    FNM: list[float] = Field(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        description="fractions of PB metabolism N. into (RPON,LPON,DON,NH4,SRPON); dim(PB=1:3,5)",
    )
    FPM: list[float] = Field(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        description="fractions of PB metabolism P. into (RPOP,LPOP,DOP,PO4,SRPOP); dim(PB=1:3,5)",
    )
    Nit: float = Field(0.07, description="maximum nitrification rate (day-1)")
    TNit: float = Field(27.0, description="optimal temp. for nitrification (oC)")
    KTNit: list[float] = Field(
        [0.0045, 0.0045],
        description="temp. dependence (T<=TNit & T>TNit) for nitrification (oC-2)",
    )
    KhDOn: float = Field(1.0, description="DO half saturation for nitrification (mg/L)")
    KhDOox: float = Field(
        0.5,
        description="DO half saturation for dentrification & DOC's oxic respiration (mg/L)",
    )
    KhNO3dn: float = Field(
        0.1, description="NO3 half saturation for denitrification (mg/L)"
    )
    KC0: list[float] = Field(
        [0.005, 0.075, 0.2], description="minimum decay rate of RPOC,LPOC,DOC (day-1)"
    )
    KN0: list[float] = Field(
        [0.005, 0.075, 0.2], description="minimum decay rate of RPON,LPON,DON (day-1)"
    )
    KP0: list[float] = Field(
        [0.005, 0.075, 0.2], description="minimum decay rate of RPOP,LPOP,DOP (day-1)"
    )
    KCalg: list[float] = Field(
        [0.0, 0.0, 0.0],
        description="algae effect on RPOC,LPOC,DOC decay (day-1.m3.g[C]-1)",
    )
    KNalg: list[float] = Field(
        [0.0, 0.0, 0.0],
        description="algae effect on RPON,LPON,DON decay (day-1.m3.g[C]-1)",
    )
    KPalg: list[float] = Field(
        [0.0, 0.0, 0.0],
        description="algae effect on RPOP,LPOP,DOP decay (day-1.m3.g[C]-1)",
    )
    TRM: list[float] = Field(
        [20.0, 20.0, 20.0], description="reference temp. for (RPOM,LPOM,DOM) decay (oC)"
    )
    KTRM: list[float] = Field(
        [0.069, 0.069, 0.069],
        description="temp. dependence for (RPOM,LPOM,DOM) decay (oC-1)",
    )
    KSR0: list[float] = Field(
        [0.001, 0.001, 0.001], description="decay rates of SRPOC,SRPON,SRPOP (day-1)"
    )
    TRSR: list[float] = Field(
        [20.0, 20.0, 20.0],
        description="reference temp. for (SRPOC,SRPON,SRPOP) decay (oC)",
    )
    KTRSR: list[float] = Field(
        [0.069, 0.069, 0.069],
        description="temp. dependence for (SRPOC,SRPON,SRPOP) decay (oC-1)",
    )
    KPIP: float = Field(0.0, description="dissolution rate of PIP (day-1)")
    KCD: float = Field(1.0, description="oxidation rate of COD at TRCOD (day-1)")
    TRCOD: float = Field(20.0, description="reference temp. for COD oxidation (oC)")
    KTRCOD: float = Field(
        0.041, description="temp. dependence for COD oxidation (oC-1)"
    )
    KhCOD: float = Field(
        1.5, description="COD half saturation for COD oxidation (mg[O2]/L)"
    )
    KhN: list[float] = Field(
        [0.01, 0.01, 0.01], description="nitrogen half saturation (mg/L)"
    )
    KhP: list[float] = Field([0.001, 0.001, 0.001], description="ph")


class SFM(RompyBaseModel):
    """
    Sediment flux model (SFM) parameter
    """

    btemp0: float = Field(5.0, description="Initial temperature (oC)")
    bstc0: float = Field(0.1, description="Surface transfer coefficient")
    bSTR0: float = Field(0.0, description="Benthic stress (day)")
    bThp0: float = Field(0.0, description="Consecutive days of hypoxia (day)")
    bTox0: float = Field(
        0.0, description="Consecutive days of oxic condition after hypoxia event (day)"
    )
    bNH40: float = Field(4.0, description="NH4")
    bNO30: float = Field(1.0, description="NO3")
    bPO40: float = Field(5.0, description="PO4")
    bH2S0: float = Field(250.0, description="H2S")
    bCH40: float = Field(40.0, description="CH4")
    bPOS0: float = Field(500.0, description="POS")
    bSA0: float = Field(500.0, description="SA")
    bPOC0: list[float] = Field([1000.0, 3000.0, 5000.0], description="POC (G=1:3)")
    bPON0: list[float] = Field([150.0, 500.0, 1500.0], description="PON (G=1:3)")
    bPOP0: list[float] = Field([30.0, 300.0, 500.0], description="POP (G=1:3)")
    bdz: float = Field(0.1, description="Sediment thickness (m)")
    bVb: float = Field(
        1.37e-5, description="Burial rate (m.day-1); 1 cm/yr=2.74e-5 m.day-1"
    )
    bsolid: list[float] = Field(
        [0.5, 0.5], description="Sediment solid conc. in Layer 1 and Layer 2 (Kg.L-1)"
    )
    bdiff: float = Field(
        1.8e-7, description="Diffusion coefficient for sediment temp. (m2/s)"
    )
    bTR: int = Field(20, description="Reference temp. for sediment processes")
    bVpmin: float = Field(
        3.0e-6, description="Minimum particle mixing velocity coefficient (m.day-1)"
    )
    bVp: float = Field(
        1.2e-4, description="Particle mixing velocity coefficient (m.day-1)"
    )
    bVd: float = Field(1.0e-3, description="Diffusion velocity coefficient (m.day-1)")
    bKTVp: float = Field(
        1.117, description="Temp. dependence of particle mixing velocity"
    )
    bKTVd: float = Field(1.08, description="Temp. dependence of diffusion velocity")
    bKST: float = Field(
        0.03, description="1st order decay rate of benthic stress (day-1)"
    )
    bSTmax: float = Field(
        20.0,
        description="Maximum value of benthic stress (day) (note: smaller than 1/bKST)",
    )
    bKhDO_Vp: float = Field(
        4.0, description="DO half-saturation of particle mixing (mg/L)"
    )
    bDOc_ST: float = Field(1.0, description="DO criteria for benthic stress (mg/L)")
    banoxic: float = Field(
        10.0,
        description="Consecutive days of hypoxia causing maximum benthic stress (day)",
    )
    boxic: float = Field(
        45.0, description="Time lag for bethos recovery from hypoxia event (day)"
    )
    bp2d: float = Field(
        0.0,
        description="Ratio from mixing coef. to diffusion coef. (benthos enhanced effect)",
    )
    bKC: list[float] = Field(
        [0.035, 0.0018, 0.00], description="Decay rate of POC (3G class) at bTR (day-1)"
    )
    bKN: list[float] = Field(
        [0.035, 0.0018, 0.00], description="Decay rate of PON (3G class) at bTR (day-1)"
    )
    bKP: list[float] = Field(
        [0.035, 0.0018, 0.00], description="Decay rate of POP (3G class) at bTR (day-1)"
    )
    bKTC: list[float] = Field(
        [1.10, 1.150, 1.17], description="Temp. dependence of POC decay (oC-1)"
    )
    bKTN: list[float] = Field(
        [1.10, 1.150, 1.17], description="Temp. dependence of PON decay (oC-1)"
    )
    bKTP: list[float] = Field(
        [1.10, 1.150, 1.17], description="Temp. dependence of POP decay (oC-1)"
    )
    bFCP: list[float] = Field(
        [0.35, 0.55, 0.01, 0.35, 0.55, 0.01, 0.35],
        description="fractions of benthic carbon into (RPOC,LPOC,DOC,SRPOC); dim(G=1:3,4)",
    )


class Silica(RompyBaseModel):
    FSP: list[float] = Field(
        [0.90, 0.10], description="Fractions of diatom silica into (SU,SA)"
    )
    FSM: list[float] = Field(
        [0.50, 0.50], description="Fractions of diatom metabolism Si into (SU,SA)"
    )
    KS: float = Field(0.03, description="Dissolution rate of SU at TRS (day-1)")
    TRS: float = Field(20.0, description="Reference temp. for SU dissolution (oC)")
    KTRS: float = Field(0.092, description="Temp. dependence for SU dissolution (oC-1)")
    KhS: list[float] = Field(
        [0.05, 0.0, 0.0],
        description="Silica half saturation (mg/L); (0.0: no Si limitation)",
    )
    s2c: list[float] = Field(
        [0.50, 0.0, 0.0],
        description="Silica to carbon ratio for phytolankton; (0.0: no Si uptake)",
    )
    KSAp: float = Field(
        0.0, description="Coefficient relating Silicate(SA) sorption to TSS"
    )


class ZB(RompyBaseModel):
    """
    Zooplankton (ZB) parameters
    ZB prey(1:8)=(ZB1,ZB2,PB1,PB2,PB3,RPOC,LPOC,DOC)
    """

    zGPM: list[float] = Field(
        [
            0.0,
            0.0,
            1.75,
            1.75,
            1.75,
            1.75,
            1.75,
            1.75,
            1.0,
            0.0,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
        ],
        description="ZB predation rate (day-1); dim(prey=1:8, ZB=1:2)",
    )
    zKhG: list[float] = Field(
        [
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
            0.175,
        ],
        description="reference prey conc. (mg/L); dim(prey=1:8, ZB=1:2)",
    )
    zTGP: list[float] = Field(
        [25.0, 25.0], description="optimal temp. for ZB growth (oC)"
    )
    zKTGP: list[float] = Field(
        [0.0035, 0.008, 0.025, 0.030],
        description="temp. dependence for ZB growth (T<=zTGP & T>zTGP); dim(ZB=1:2,1:2) (oC-2)",
    )
    zAG: float = Field(0.75, description="ZB assimilation efficiency ratio (0-1)")
    zRG: float = Field(0.1, description="ZB respiration ratio when it grazes (0-1)")
    zMRT: list[float] = Field([0.02, 0.02], description="ZB mortality rates (day-1)")
    zMTB: list[float] = Field([0.254, 0.186], description="ZB metabolism rates (day-1)")
    zTMT: list[float] = Field(
        [20.0, 20.0], description="reference temp. for ZB metabolism (oC)"
    )
    zKTMT: list[float] = Field(
        [0.0693, 0.0693], description="temp. dependence for ZB metabolism (oC-1)"
    )
    zFCP: list[float] = Field(
        [0.35, 0.55, 0.10], description="fractions of ZB carbon into (RPOC,LPOC,DOC)"
    )
    zFNP: list[float] = Field(
        [0.35, 0.50, 0.10, 0.05],
        description="fractions of ZB nitrogen into (RPON,LPON,DON,NH4)",
    )
    zFPP: list[float] = Field(
        [0.10, 0.20, 0.50, 0.20],
        description="fractions of ZB Phosphorus into (RPOP,LPOP,DOP,PO4)",
    )
    zFSP: list[float] = Field(
        [0.70, 0.25], description="fractions of ZB silica into (SU,SA)"
    )
    zFCM: list[float] = Field(
        [0.10, 0.0],
        description="fractions of ZB metabolism carbon into DOC; dim(ZB=1:2)",
    )
    zFNM: list[float] = Field(
        [0.35, 0.30, 0.50, 0.40, 0.10, 0.20, 0.05, 0.10],
        description="fractions of ZB metabolism N. into (RPON,LPON,DON,NH4); dim(ZB=1:2,4)",
    )
    zFPM: list[float] = Field(
        [0.35, 0.30, 0.50, 0.40, 0.10, 0.20, 0.05, 0.10],
        description="fractions of ZB metabolism P. into (RPOP,LPOP,DOP,PO4); dim(ZB=1:2,4)",
    )
    zFSM: list[float] = Field(
        [0.50, 0.40, 0.50, 0.60],
        description="fractions of ZB metabolism Si into (SU,SA); dim(ZB=1:2,2)",
    )
    zKhDO: list[float] = Field(
        [0.5, 0.5], description="DO half saturation for ZB's DOC excretion (mg/L)"
    )
    zn2c: list[float] = Field(
        [0.20, 0.20], description="nitrogen to carbon ratio for zooplankton"
    )
    zp2c: list[float] = Field(
        [0.02, 0.02], description="phosphorus to carbon ratio for zooplankton"
    )
    zs2c: list[float] = Field(
        [0.50, 0.50], description="silica to carbon ratio for zooplankton"
    )
    z2pr: list[float] = Field(
        [0.5, 0.5],
        description="ratio converting ZB+PB biomass to predation rates on ZB (L.mg-1.day-1)",
    )
    p2pr: float = Field(
        0.25,
        description="ratio converting ZB+PB biomass to predation rates on PB (L.mg-1.day-1)",
    )


class PH_ICM(RompyBaseModel):
    """
    PH model parameter
    """

    ppatch0: int = Field(
        -999, description="Region flag for pH (1: ON all elem.; -999: spatial)"
    )
    pKCACO3: float = Field(60.0, description="Dissolution between CaCO3 and Ca++")
    pKCA: float = Field(
        60.0, description="Sediment surface transfer coefficient from CaCO3 to Ca++"
    )
    pRea: float = Field(1.0, description="Reaeration rate for CO2")
    inu_ph: int = Field(0, description="Nudge option for pH model")


class SAV(RompyBaseModel):
    """
    Submerged Aquatic Vegetation (SAV) parameters
    """

    spatch0: int = Field(
        -999, description="Region flag for SAV. (1: ON all elem.; -999: spatial)"
    )
    stleaf0: int = Field(-999, description="Initial concentration of total SAV leaf")
    ststem0: int = Field(-999, description="Initial concentration of total SAV stem")
    stroot0: int = Field(-999, description="Initial concentration of total SAV root")
    sGPM: float = Field(0.1, description="Maximum growth rate (day-1)")
    sTGP: int = Field(32, description="Optimal growth temperature (oC)")
    sKTGP: str = Field(
        "0.003  0.005",
        description="Temperature dependence for growth (T<=sTGP & T>sTGP)",
    )
    sFAM: float = Field(
        0.2, description="Fraction of leaf production to active metabolism"
    )
    sFCP: str = Field(
        "0.6    0.3    0.1",
        description="Fractions of production to leaf/stem/root biomass",
    )
    sMTB: str = Field(
        "0.02   0.02   0.02", description="Metabolism rates of leaf/stem/root"
    )
    sTMT: str = Field(
        "20     20     20", description="Reference temp. for leaf/stem/root metabolism"
    )
    sKTMT: str = Field(
        "0.069  0.069  0.069",
        description="Temperature dependence of leaf/stem/root metabolism",
    )
    sFCM: str = Field(
        "0.05   0.15   0.3   0.5",
        description="Fractions of metabolism C into (RPOC,LPOC,DOC,CO2)",
    )
    sFNM: str = Field(
        "0.05   0.15   0.3   0.5",
        description="Fractions of metabolism N into (RPON,LPON,DON,NH4)",
    )
    sFPM: str = Field(
        "0.05   0.1    0.35  0.5",
        description="Fractions of metabolism P into (RPOP,LPOP,DOP,PO4)",
    )
    sKhNw: float = Field(0.01, description="Nitrogen half saturation in water column")
    sKhNs: float = Field(0.1, description="Nitrogen half saturation in sediments")
    sKhNH4: float = Field(0.1, description="Ammonium half saturation")
    sKhPw: float = Field(
        0.001, description="Phosphorus half saturation in water column"
    )
    sKhPs: float = Field(0.01, description="Phosphorus half saturation in sediments")
    salpha: float = Field(
        0.006, description="Initial slope of P-I curve (g[C]*m2/g[Chl]/E)"
    )
    sKe: float = Field(0.045, description="Light attenuation due to SAV absorption")
    shtm: str = Field(
        "0.054  2.0", description="Minimum (base) and maximum canopy height"
    )
    s2ht: str = Field(
        "0.0036 0.0036 0.0",
        description="Coefficients converting (leaf,stem,root) to canopy height",
    )
    sc2dw: float = Field(0.38, description="Carbon to dry weight ratio of SAV")
    s2den: int = Field(
        10, description="Coefficient computing SAV density from leaf&stem"
    )
    sn2c: float = Field(0.09, description="Nitrogen to carbon ratio of SAV")
    sp2c: float = Field(0.01, description="Phosphorus to carbon ratio")
    so2c: float = Field(2.67, description="Oxygen to carbon ratio")


# &VEG
# !-----------------------------------------------------------------------
# !Intertidal Vegetation (VEG), paramters
# !-----------------------------------------------------------------------


class VEG(RompyBaseModel):
    """
    Intertidal Vegetation (VEG), paramters
    """

    vpatch0: int = Field(
        -999, description="Region flag for VEG. (1: ON all elem.; -999: spatial)"
    )
    vtleaf0: list[float] = Field(
        [100.0, 100.0, 100.0], description="Init conc to total veg leaf"
    )
    vtstem0: list[float] = Field(
        [100.0, 100.0, 100.0], description="Init conc to total veg stem"
    )
    vtroot0: list[float] = Field(
        [30.0, 30.0, 30.0], description="Init conc to total veg root"
    )
    vGPM: list[float] = Field(
        [0.1, 0.1, 0.1], description="Maximum growth rate (day-1)"
    )
    vFAM: list[float] = Field(
        [0.2, 0.2, 0.2], description="Fractions of leaf production to active metabolism"
    )
    vTGP: list[float] = Field(
        [32.0, 32.0, 32.0], description="Optimal growth temperature (oC)"
    )
    vKTGP: list[float] = Field(
        [0.003, 0.003, 0.003, 0.005, 0.005, 0.005],
        description="Temp. dependence(3,2) for growth (T<=vTGP & T>vTGP)",
    )
    vFCP: list[float] = Field(
        [0.6, 0.6, 0.6, 0.3, 0.3, 0.3, 0.1, 0.1, 0.1],
        description="Fractions of production to leaf/stem/root biomass; dim(veg,leaf/stem/root)",
    )
    vMTB: list[float] = Field(
        [0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.01, 0.01, 0.01],
        description="Metabolism rates of leaf/stem/root; dim(veg,leaf/stem/root)",
    )
    vTMT: list[float] = Field(
        [20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0],
        description="Reference temp. for leaf/stem/root metabolism",
    )
    vKTMT: list[float] = Field(
        [0.069, 0.069, 0.069, 0.069, 0.069, 0.069, 0.069, 0.069, 0.069],
        description="Temp. dependence for leaf/stem/root metabolism",
    )
    vFNM: list[float] = Field(
        [0.05, 0.05, 0.05, 0.15, 0.15, 0.15, 0.3, 0.3, 0.3, 0.5, 0.5, 0.5],
        description="Fractions(3,4) of metabolism N into (RPON,LPON,DON,NH4)",
    )
    vFPM: list[float] = Field(
        [0.05, 0.05, 0.05, 0.1, 0.1, 0.1, 0.35, 0.35, 0.35, 0.5, 0.5, 0.5],
        description="Fractions(3,4) of metabolism P into (RPOP,LPOP,DOP,PO4)",
    )
    vFCM: list[float] = Field(
        [0.05, 0.05, 0.05, 0.15, 0.15, 0.15, 0.3, 0.3, 0.3, 0.5, 0.5, 0.5],
        description="Fractions(3,4) of metabolism N into (RPOC,LPOC,DOC,CO2)",
    )
    ivNc: int = Field(1, description="Recycled veg N goes to (0: sediment; 1: water)")
    ivPc: int = Field(1, description="Recycled veg P goes to (0: sediment; 1: water)")
    vKhNs: list[float] = Field(
        [0.1, 0.1, 0.1], description="Nitrogen half saturation in sediments"
    )
    vKhPs: list[float] = Field(
        [0.01, 0.01, 0.01], description="Phosphorus half saturation in sediments"
    )
    vScr: float = Field(35.0, description="Reference sality for computing veg growth")
    vSopt: list[float] = Field(
        [35.0, 15.0, 0.0], description="Optimal salinity for veg growth"
    )
    vInun: list[float] = Field(
        [1.0, 1.0, 1.0],
        description="Reference value for inundation stress (nondimensional)",
    )
    ivNs: int = Field(1, description="N limitation on veg growth(0: OFF; 1: ON)")
    ivPs: int = Field(1, description="P limitation on veg growth(0: OFF; 1: ON)")
    ivMRT: int = Field(0, description="Veg mortality term (0: OFF;  1: ON)")
    vTMR: list[float] = Field(
        [17.0, 17.0, 17.0, 17.0, 17.0, 17.0],
        description="Reference temp(3,2) for leaf/stem mortality",
    )
    vKTMR: list[float] = Field(
        [4.0, 4.0, 4.0, 4.0, 4.0, 4.0],
        description="Temp dependence(3,2) for leaf/stem mortality",
    )
    vMR0: list[float] = Field(
        [12.8, 12.8, 12.8, 12.8, 12.8, 12.8],
        description="Base value(3,2) of temp effect on mortality (unit: None)",
    )
    vMRcr: list[float] = Field(
        [15.0, 15.0, 15.0, 15.0, 15.0, 15.0],
        description="Reference value(3,2) for computing mortality (unit: None)",
    )
    valpha: list[float] = Field(
        [0.006, 0.006, 0.006], description="Init. slope of P-I curve"
    )
    vKe: list[float] = Field(
        [0.045, 0.045, 0.045], description="Light attenuation from veg absorption"
    )
    vht0: list[float] = Field(
        [0.054, 0.054, 0.054], description="Base veg canopy hgt (vht)"
    )
    vcrit: list[float] = Field(
        [250.0, 250.0, 250.0], description="Critical mass for computing vht"
    )
    v2ht: list[float] = Field(
        [0.0036, 0.0036, 0.0036, 0.001, 0.001, 0.001],
        description="Coefs. convert mass(3,2) to canopy height",
    )
    vc2dw: list[float] = Field(
        [0.38, 0.38, 0.38], description="Carbon to dry weight ratio of VEG"
    )
    v2den: list[int] = Field([10, 10, 10], description="Coeff. computing veg density")
    vp2c: list[float] = Field(
        [0.01, 0.01, 0.01], description="Phosphorus to carbon ratio"
    )
    vn2c: list[float] = Field(
        [0.09, 0.09, 0.09], description="Nitrogen to carbon ratio"
    )
    vo2c: list[float] = Field([2.67, 2.67, 2.67], description="Oxygen to carbon ratio")


class BAG(RompyBaseModel):
    """
    Benthic Algae
    """

    gpatch0: int = Field(
        -999, description="Region flag for BA. (1: ON all elem.; -999: spatial)"
    )
    BA0: float = Field(5.0, description="Initial BA concentration (g[C].m-2)")
    gGPM: float = Field(2.25, description="BA maximum growth rate (day-1)")
    gTGP: float = Field(20.0, description="Optimal temp. for BA growth (oC)")
    gKTGP: str = Field(
        "0.004  0.006", description="Temp. dependence for BA growth (oC-2)"
    )
    gMTB: float = Field(0.05, description="Respiration rate at temp. of gTR (day-1)")
    gPRR: float = Field(0.1, description="Predation rate at temp. of gTR (day-1)")
    gTR: float = Field(
        20.0, description="Reference temperature for BA respiration (oC)"
    )
    gKTR: float = Field(0.069, description="Temp. dependence for BA respiration (oC-1)")
    galpha: float = Field(0.1, description="Init. slope of P-I curve (m2/E)")
    gKSED: float = Field(0.0, description="Light attenuation due to sediment (None)")
    gKBA: float = Field(
        0.01, description="Light attenuation coef. due to BA self-shading (g[C]-1.m2)"
    )
    gKhN: float = Field(
        0.01, description="Nitrogen half saturation for BA growth (g[N]/m2)"
    )
    gKhP: float = Field(
        0.001, description="Phosphorus half saturation for BA growth (g[P]/2)"
    )
    gp2c: float = Field(0.0167, description="Phosphorus to carbon ratio")
    gn2c: float = Field(0.167, description="Nitrogen to carbon ratio")
    go2c: float = Field(2.67, description="Oxygen to carbon ratio")
    gFCP: str = Field(
        "0.5 0.45 0.05",
        description="Fraction of predation BA C into 3G classes in sediment",
    )
    gFNP: str = Field(
        "0.5 0.45 0.05",
        description="Fraction of predation BA N into 3G classes in sediment",
    )
    gFPP: str = Field(
        "0.5 0.45 0.05",
        description="Fraction of predation BA P into 3G classes in sediment",
    )


class ERO(RompyBaseModel):
    """
    Benthic erosion parameter
    """

    ierosion: int = Field(0, description="1: H2S flux; 2: POC flux; 3: H2S&POC flux")
    erosion: int = Field(864, description="erosion rate (kg.m-2.day)")
    etau: float = Field(1e-6, description="critical bottom shear stress (Pa)")
    eporo: float = Field(
        0.8, description="coefficient in erosion formula (see code in icm_sfm.F90)"
    )
    efrac: float = Field(
        0.5,
        description="fraction coefficient in erosion formula (see code in icm_sfm.F90)",
    )
    ediso: float = Field(
        2.5, description="H2S erosion coefficient (see code in icm_sfm.F90)"
    )
    dfrac: str = Field(
        "0.02 0.02",
        description="deposition fraction of POC (negative value: dfrac will be computed)",
    )
    dWS_POC: str = Field(
        "3.0 3.0", description="coefficient in POC erosion (see code in icm_sfm.F90)"
    )


class ICM(RompyBaseModel):
    marco: MARCO = Field(MARCO())
    sfm: SFM = Field(SFM())
    silica: Silica = Field(Silica())
    zb: ZB = Field(ZB())
    ph_icm: PH_ICM = Field(PH_ICM())
    sav: SAV = Field(SAV())
    bag: BAG = Field(BAG())
    ero: ERO = Field(ERO())
    veg: VEG = Field(VEG())



import os

from pydantic import Field

from rompy_schism.namelists.basemodel import NamelistBaseModel


class Sflux_Inputs(NamelistBaseModel):
    air_1_relative_weight: float = Field(
        1.0,
        description="air_[12]_relative_weight set the relative ratio between datasets '1' and '2'",
    )
    air_2_relative_weight: float = Field(
        99.0,
        description="air_[12]_relative_weight set the relative ratio between datasets '1' and '2'",
    )
    air_1_max_window_hours: float = Field(
        120.0,
        description="max. # of hours (offset from start time in each file) in each file of set '1'",
    )
    air_2_max_window_hours: float = Field(
        120.0,
        description="max. # of hours (offset from start time in each file) in each file of set '1'",
    )
    air_1_fail_if_missing: bool = Field(True, description="set '1' is mandatory")
    air_2_fail_if_missing: bool = Field(False, description="set '2' is optional")
    air_1_file: str = Field("air_1", description="file name for 1st set of 'air'")
    air_2_file: str = Field("air_2", description="file name for 2nd set of 'air'")
    uwind_name: str = Field("uwind", description="name of u-wind vel.")
    vwind_name: str = Field("vwind", description="name of v-wind vel.")
    prmsl_name: str = Field(
        "prmsl", description="name of air pressure (@MSL) variable in .nc file"
    )
    stmp_name: str = Field("stmp", description="name of surface air T")
    spfh_name: str = Field("spfh", description="name of specific humidity")
    rad_1_relative_weight: float = Field(1.0)
    rad_2_relative_weight: float = Field(99.0)
    rad_1_max_window_hours: float = Field(24.0)
    rad_2_max_window_hours: float = Field(24.0)
    rad_1_fail_if_missing: bool = Field(False)
    rad_2_fail_if_missing: bool = Field(False)
    rad_1_file: str = Field("rad_1")
    rad_2_file: str = Field("rad_2")
    dlwrf_name: str = Field("dlwrf")
    dswrf_name: str = Field("dswrf")
    prc_1_relative_weight: float = Field(1.0)
    prc_2_relative_weight: float = Field(99.0)
    prc_1_max_window_hours: float = Field(24.0)
    prc_2_max_window_hours: float = Field(24.0)
    prc_1_fail_if_missing: bool = Field(False)
    prc_2_fail_if_missing: bool = Field(False)
    prc_1_file: str = Field("prc_1")
    prc_2_file: str = Field("prc_2")
    prate_name: str = Field("prate", description="name of precipitation rate variable")
    priority: int = Field(1, description="priority of the data")

    def render(self) -> str:
        """Render the namelist variable as a string"""
        # create string of the form "variable = value"
        ret = []
        ret += ["! SCHISM sflux namelist rendered from Rompy"]
        ret += ["&sflux_inputs"]
        for key, value in self.model_dump().items():
            if key in ["id", "priority", "file_number"]:
                continue
            if isinstance(value, list):
                value = ", ".join([str(item) for item in value])
            if isinstance(value, str):
                value = f"'{value}'"
            if isinstance(value, bool):
                value = ".true." if value else ".false."
            ret += [f"{key}={value},"]
        ret += ["/\n"]
        return "\n".join(ret)

    def write_nml(self, destdir: str) -> None:
        """Write the namelist to a file"""
        os.makedirs(destdir, exist_ok=True)
        output = destdir / f"{self.__class__.__name__.lower()}.txt"
        with open(output, "w") as f:
            f.write(self.render())
        return output
